"""
Aerial refueling double tank balancing example.

The system consists of two fuel tanks, T1 and T2.
Fuel is drawn at a constant rate from T2.
There are two dynamic modes:

    - 'normal operation' and
    - 'air refuel mode'.

In 'air refuel mode' fuel is added
to T1 at a constant rate.

The objective is to keep the fuel levels in
the tanks close to each other.

reference
=========
    Nilsson P.; Ozay N.; Topcu U.; Murray R.M.
     Temporal Logic Control of Switched Affine Systems
     with an Application in Fuel Balancing
    2012 American Control Conference
"""

import os
import pickle
import numpy as np

from tulip import hybrid, abstract, spec, synth
import polytope as pc
from tulip.abstract.plot import plot_strategy

import sys
sys.path.append('../')
import tomatlab

"""Problem variables"""
tank_capacity = 10      # Maximum tank capacity
fuel_consumption = 1    # Rate at which fuel is drawn from tank 2
refill_rate = 3         # Rate at which fuel is refilled in tank 1 during refuel mode
input_lb = 0            # Lower bound on fuel move rate from tank 1 to 2
input_ub = 3            # Upper bound on fuel move rate from tank 1 to 2
max_vol_diff = 2        # Enforce |v1-v2| to be lower than this value
fin_vol_diff = 1        # Enforce |v1-v2| to alw. ev. be lower than this value
max_refuel_level = 8    # Above this level no refueling will take place
N = 1                   # Horizon length
disturbance = 0.0       # Absolute uncertainty in fuel consumption
init_lower = 6          # Lower bound for possible initial volumes
init_upper = 8          # Upper bound for possible initial volumes


"""State space and propositions"""
cont_ss = pc.Polytope(
    np.array([[1,0],
              [-1,0],
              [0,1],
              [0,-1]]),
    np.array([tank_capacity, 0, tank_capacity, 0])
)

cont_props = {}
cont_props['no_refuel'] = pc.Polytope(
    np.array([[1,0],
              [-1,0],
              [0,1],
              [0,-1]]),
    np.array([tank_capacity, 0, tank_capacity, -max_refuel_level])
)
cont_props['vol_diff'] = pc.Polytope(
    np.array([[-1,0],
              [0,-1],
              [-1,1],
              [1,-1]]),
    np.array([0,0,max_vol_diff,max_vol_diff])
)
cont_props['initial'] = pc.Polytope(
    np.array([[1,0],
             [-1,0],
             [0, 1],
             [0,-1]]),
    np.array([init_upper, -init_lower, init_upper, -init_lower])
)
cont_props['critical'] = pc.Polytope(
    np.array([[-1,0],
              [0,-1],
              [1,1]]),
    np.array([0, 0, 2*fuel_consumption])
)

"""Dynamics"""
A = np.eye(2)
B = np.array([[-1],[1]])
E = np.array([[0],[1]])
K1 = np.array([[0.],[-fuel_consumption]])
K2 = np.array([[refill_rate],[-fuel_consumption]])
U1 = pc.Polytope(
    np.array([
        [1,0,0],
        [-1,0,0],
        [1,-1,0]
    ]),
    np.array([input_ub, -input_lb, 0])
)
U2 = pc.Polytope(
    np.array([
        [1,0,0],
        [-1,0,0],
        [1,-1,0]
    ]),
    np.array([input_ub,-input_lb, refill_rate])
)
W = pc.Polytope(np.array([[1],[-1]]),
                np.array([disturbance, disturbance]))

# Normal operation dynamics
cont_dyn_normal = hybrid.LtiSysDyn(A, B, E, K1, U1, W, domain=cont_ss)

# Aerial refueling mode dynamics
cont_dyn_refuel = hybrid.LtiSysDyn(A, B, E, K2, U2, W, domain=cont_ss)

"""Switched Dynamics"""
env_modes = ('normal', 'refuel')
sys_modes = ('fly',)

pwa_normal = hybrid.PwaSysDyn([cont_dyn_normal], domain=cont_ss)
pwa_refuel = hybrid.PwaSysDyn([cont_dyn_refuel], domain=cont_ss)

dynamics_dict = {
    ('normal', 'fly') : pwa_normal,
    ('refuel', 'fly') : pwa_refuel
}

switched_dynamics = hybrid.SwitchedSysDyn(
    cts_ss=cont_ss,
    disc_domain_size=(len(env_modes), len(sys_modes)),
    dynamics=dynamics_dict,
    env_labels=env_modes,
    disc_sys_labels=sys_modes
)

"""Create convex proposition preserving partition"""
ppp = abstract.prop2part(cont_ss, cont_props)
ppp, new2old = abstract.part2convex(ppp)


"""Discretize to establish transitions"""

disc_params = {}
disc_params[('normal', 'fly')] = {'N':N, 'trans_length':3}
disc_params[('refuel', 'fly')] = {'N':N, 'trans_length':3}

sys_ts = abstract.discretize_switched(
    ppp, switched_dynamics, disc_params, plot=False
)

"""Specs"""
env_vars = set()
sys_disc_vars = set()

env_init = {'env_actions = normal'}

env_safe = {'no_refuel -> X(env_actions = normal)',
            '(critical & (env_actions = normal)) -> X(env_actions = refuel)',
            '(!critical & env_actions = normal) -> X(env_actions = normal)',
            '(!no_refuel & env_actions = refuel) -> X(env_actions = refuel)'}
env_prog = {'env_actions = refuel'}

# relate switching actions to u_in (env_actions)
sys_init = {'initial'}
sys_safe = {'vol_diff'}
sys_prog = {'True'}

specs = spec.GRSpec(env_vars, sys_disc_vars,
                    env_init, sys_init,
                    env_safe, sys_safe,
                    env_prog, sys_prog)
print(specs.pretty())

"""Synthesis"""
print("Starting synthesis")

ctrl = synth.synthesize(
    'gr1c', specs, sys=sys_ts.ts, ignore_sys_init=True,
)

disc_params = disc_params[('normal', 'fly')]
tomatlab.export('fuel_tank.mat', ctrl, switched_dynamics, sys_ts, disc_params)



#!/usr/bin/env python
"""
Tests for the interface with gr1py.
"""
import logging
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('tulip.spec.lexyacc').setLevel(logging.WARNING)
from nose.tools import assert_raises
import os
from tulip.spec import GRSpec, translate
from tulip.interfaces import gr1py


class basic_test(object):
    def setUp(self):
        self.f_un = GRSpec(
            env_vars="x",
            sys_vars="y",
            env_init="x",
            env_prog="x",
            sys_init="y",
            sys_safety=["y -> X(!y)", "!y -> X(y)"],
            sys_prog="y && x",
            moore=False,
            plus_one=False,
            qinit='\A \E')
        self.dcounter = GRSpec(
            sys_vars={"y": (0, 5)},
            sys_init=["y=0"],
            sys_prog=["y=0", "y=5"],
            moore=False,
            plus_one=False,
            qinit='\A \E')

    def tearDown(self):
        self.f_un = None
        self.dcounter = None

    def test_check_realizable(self):
        # f_un
        assert not gr1py.check_realizable(self.f_un)
        self.f_un.sys_safety = []
        assert gr1py.check_realizable(self.f_un)
        self.f_un.qinit = '\A \A'
        self.f_un.env_init = ['x & y']
        self.f_un.sys_init = list()
        with assert_raises(AssertionError):
            assert gr1py.check_realizable(self.f_un)
        # counter
        assert gr1py.check_realizable(self.dcounter)
        self.dcounter.qinit = '\A \A'
        self.dcounter.sys_init = list()
        with assert_raises(AssertionError):
            assert gr1py.check_realizable(self.dcounter)

    def test_synthesize(self):
        self.f_un.sys_safety = []  # Make it realizable
        g = gr1py.synthesize(self.f_un)
        assert g is not None
        assert len(g.env_vars) == 1 and 'x' in g.env_vars
        assert len(g.sys_vars) == 1 and 'y' in g.sys_vars

        g = gr1py.synthesize(self.dcounter)
        assert g is not None
        assert len(g.env_vars) == 0
        assert len(g.sys_vars) == 1 and 'y' in g.sys_vars
        assert len(g) == 2, [g.nodes(data=True), g.edges(data=True)]



# Copyright (c) 2011-2014 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""
Classes representing hybrid dynamical systems.
"""
from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

from warnings import warn
import itertools
from pprint import pformat

import numpy as np
import polytope as pc

# inline imports:
#
# from tulip.graphics import newax, quiver

def _indent(s, n):
    s = s.split('\n')
    w = n*' '
    return w + ('\n'+w).join(s)

class LtiSysDyn(object):
    """Represent discrete-time continuous-state dynamics::

        s[t+1] = A*s[t] + B*u[t] + E*d[t] + K

    subject to the constraints::

        u[t] \in Uset
        d[t] \in Wset
        s[t] \in domain

    where:
        - u[t] the control input
        - d[t] the disturbance input
        - s[t] the system state

    A LtiSysDyn object contains the fields:

        - A, B, E, K, (matrices)
        - Uset, Wset, (each a C{polytope.Polytope})
        - domain (C{polytope.Polytope} or C{polytope.Region})
        - time_semantics: 'discrete' (if system is originally a discrete-time
          system) or 'sampled' (if system is sampled from a continuous-time
          system)
        - timestep: A positive real number containing the timestep (for sampled
          system)

    as defined above.

    Note
    ====
    For state-dependent bounds on the input,::
        [u[t]; s[t]] \in Uset
    can be used.

    See Also
    ========
    L{PwaSysDyn}, L{SwitchedSysDyn}, C{polytope.Polytope}
    """
    def __init__(self, A=None, B=None, E=None, K=None,
                 Uset=None,Wset=None, domain=None, time_semantics=None,
                 timestep=None):

        if Uset is None:
            warn('Uset not given to LtiSysDyn()')
        elif not isinstance(Uset, pc.Polytope):
            raise Exception('`Uset` has to be a Polytope')
        if domain is None:
            warn("Domain not given to LtiSysDyn()")
        if ((domain is not None) and
            (not (isinstance(domain, pc.Polytope) or
                isinstance(domain, pc.Region))
            )
        ):
            raise Exception('`domain` has to be a Polytope or Region')

        # check dimensions agree
        try:
            nA, mA = A.shape
        except:
            raise TypeError('A matrix must be 2d array')
        if nA != mA:
            raise ValueError('A must be square')
        if domain is not None:
            if domain.dim != mA:
                raise Exception('domain.dim != A.size[1]')

        if B is not None:
            try:
                nB, mB = B.shape
            except:
                raise TypeError('B matrix must be 2d array')
            if nA != nB:
                raise ValueError('A and B must have same number of rows')
            if Uset is not None:
                if (Uset.dim != mB) and (Uset.dim != mB + nA):
                    msg = 'Uset.dim != B.size[1]'
                    msg += ' and != B.size[1] + A.size[1]'
                    raise Exception(msg)

        if E is not None:
            try:
                nE, mE = E.shape
            except:
                raise TypeError('E matrix must be 2d array')
            if nA != nE:
                raise ValueError('A and E must have same number of rows')
            if Wset is not None:
                if Wset.dim != mE:
                    raise Exception('Wset.dim != E.size[1]')

        if K is not None:
            try:
                nK, mK = K.shape
            except:
                raise TypeError('K column vector must be 2d array')

            if nA != nK:
                raise ValueError('A and K must have same number of rows')
            if mK != 1:
                raise ValueError('K must be a column vector')

        self.A = A
        self.B = B

        if K is None:
            if len(A) != 0:
                self.K = np.zeros([mA, 1])
            else:
                self.K = K
        else:
            self.K = K.reshape(K.size,1)

        if E is None and (len(A) != 0):
            self.E = np.zeros([mA, 1])
            self.Wset = pc.Polytope()
        else:
            self.E = E
            self.Wset = Wset

        self.Uset = Uset
        self.domain = domain

        # Check that timestep and semantics are valid.
        _check_time_data(time_semantics, timestep)
        self.time_semantics = time_semantics
        self.timestep = timestep


    def __str__(self):
        n = 3
        output = 'A =\n' + _indent(str(self.A), n)
        output += '\nB =\n' + _indent(str(self.B), n)
        output += '\nE =\n' + _indent(str(self.E), n)
        output += '\nK =\n' + _indent(str(self.K), n)
        output += '\nUset =\n' + _indent(str(self.Uset), n)
        output += '\nWset =\n' + _indent(str(self.Wset), n)
        return output

    def plot(self, ax=None, color=np.random.rand(3), show_domain=True,
             res=(5, 5), **kwargs):
        try:
            from tulip.graphics import newax, quiver
        except:
            logger.error('failed to import graphics')
            warn('pyvectorized not found. No plotting.')
            return

        (x, res) = pc.grid_region(self.domain, res=res)
        n = self.A.shape[0]
        DA = self.A - np.eye(n)
        v = DA.dot(x) + self.K

        if ax is None:
            ax, fig = newax()

        if show_domain:
            self.domain.plot(ax, color)

        quiver(x, v, ax, **kwargs)

        return ax

class PwaSysDyn(object):
    """PwaSysDyn class for specifying a polytopic piecewise affine system.
    A PwaSysDyn object contains the fields:

      - C{list_subsys}: list of L{LtiSysDyn}

      - C{domain}: domain over which piecewise affine system is defined,
          type: polytope.Polytope or polytope.Region

      - C{time_semantics}: 'discrete' (if system is originally a discrete-time
       system) or 'sampled' (if system is sampled from a continuous-time
       system)

      - C{timestep}: A positive real number containing the timestep (for sampled
        systems)

    For the system to be well-defined the domains of its subsystems should be
    mutually exclusive (modulo intersections with empty interior) and cover the
    domain.

    See Also
    ========
    L{LtiSysDyn}, L{SwitchedSysDyn}, C{polytope.Polytope}
    """
    def __init__(self, list_subsys=[], domain=None, time_semantics=None,
                 timestep=None, overwrite_time=True):
        """
        @type overwrite_time: bool
        @param overwrite_time: If true, then overwrites any time data in the
                               objects in C{list_subsys} with the data in
                               C{time_semantics} and C{timestep} variables.
                               Otherwise checks that the time data of the
                               objects in C{list_subsys} are consistent with
                               C{time_semantics} and C{timestep}.
        """

        if domain is None:
            warn("Domain not given to PwaSysDyn()")

        if ((domain is not None) and
            (not (isinstance(domain, pc.Polytope) or
                isinstance(domain, pc.Region))
            )
        ):
            raise Exception("PwaSysDyn: `domain` has to be a Polytope or Region")

        if len(list_subsys) > 0:
            uncovered_dom = domain.copy()
            n = list_subsys[0].A.shape[1]  # State space dimension
            m = list_subsys[0].B.shape[1]  # Input space dimension
            p = list_subsys[0].E.shape[1]  # Disturbance space dimension
            for subsys in list_subsys:
                uncovered_dom = uncovered_dom.diff(subsys.domain)
                if (n!=subsys.A.shape[1] or m!=subsys.B.shape[1] or
                    p!=subsys.E.shape[1]):
                    raise Exception("PwaSysDyn: state, input, disturbance " +
                                    "dimensions have to be the same for all " +
                                     "subsystems")
            if not pc.is_empty(uncovered_dom):
                raise Exception("PwaSysDyn: subdomains must cover the domain")
            for x in itertools.combinations(list_subsys, 2):
                if pc.is_fulldim(x[0].domain.intersect(x[1].domain) ):
                    raise Exception("PwaSysDyn: subdomains have to be mutually"+
                        " exclusive")

        self.list_subsys = list_subsys
        self.domain = domain

        # Input time semantics
        _check_time_data(time_semantics, timestep)
        if overwrite_time:
            _push_time_data(self.list_subsys, time_semantics, timestep)
        else:
            _check_time_consistency(list_subsys, time_semantics, timestep)
        self.timestep = timestep
        self.time_semantics = time_semantics


    def __str__(self):
        s = 'Piecewise-Affine System Dynamics\n'
        s += 30 * '-' + 2*'\n'

        s += 3*' ' + 'Domain:\n\n'
        s += _indent(str(self.domain), n=6) + '\n'

        for i, sys in enumerate(self.list_subsys):
            s += 3*' ' + 'Subsystem: ' + str(i) +'\n'
            s += _indent(str(sys), n=6)
        return s

    @classmethod
    def from_lti(cls, A=[], B=[], E=[], K=[],
                 Uset=None, Wset=None,domain=None):
        lti_sys = LtiSysDyn(A,B,E,K,Uset,Wset,domain)
        return cls([lti_sys], domain)

    def plot(self, ax=None, show_domain=True, **kwargs):
        try:
            from tulip.graphics import newax
        except:
            logger.error('failed to import graphics')
            return

        if ax is None:
            ax, fig = newax()

        for subsystem in self.list_subsys:
            subsystem.plot(ax, color=np.random.rand(3),
                           show_domain=show_domain, **kwargs)
        return ax

class SwitchedSysDyn(object):
    """Represent hybrid systems switching between dynamic modes.

    A C{SwitchedSysDyn} represents a system with switching modes
    that depend on both discrete:

        - n_env environment variables (uncontrolled)
        - n_sys system variables (controlled)

    A C{SwitchedSysDyn} object contains the fields:

     - C{disc_domain_size}: 2-tuple of numbers of modes
       type: (n_env, n_sys)

     - C{env_labels}: (optional) labels for discrete environment variables
       type: list of len(n_env)
       default: range(n_env)

     - C{disc_sys_labels}: (optional) labels for discrete system variables
       type: list of len(n_sys)
       default: range(n_sys)

     - C{dynamics}: mapping mode 2-tuples to active dynamics::

         (env_label, sys_label) -> PwaSysDyn

       type: dict
       default: If no env_label or sys_label passed,
       then default to int indices (i,j) L{PwaSysDyn}.

     - C{cts_ss}: continuous state space over which hybrid system is defined.
       type: C{polytope.Region}

     - C{time_semantics}: 'discrete' (if system is originally a discrete-time
       system) or 'sampled' (if system is sampled from a continuous-time
       system)

     - C{timestep}: A positive real number containing the timestep (for sampled
       systems)


    Note
    ====
    We assume that system and environment switching modes are
    independent of one another.  (Use LTL statement to make it not so.)

    See Also
    ========
    L{LtiSysDyn}, L{PwaSysDyn}, C{polytope.Region}
    """
    def __init__(self, disc_domain_size=(1,1),
                 dynamics=None, cts_ss=None,
                 env_labels=None, disc_sys_labels=None, time_semantics=None,
                 timestep=None, overwrite_time=True):
        """
        @type overwrite_time: bool
        @param overwrite_time: If true, then overwrites any time data in the
                               objects in C{list_subsys} with the data in
                               C{time_semantics} and C{timestep} variables.
                               Otherwise checks that the time data of the
                               objects in C{list_subsys} are consistent with
                               C{time_semantics} and C{timestep}.
        """

        # check that the continuous domain is specified
        if cts_ss is None:
            warn('continuous state space not given to SwitchedSysDyn')
        else:
            if not isinstance(cts_ss, (pc.Polytope, pc.Region) ):
                raise Exception('SwitchedSysDyn: ' +
                   '`cts_ss` must be a Polytope or Region')

        self.disc_domain_size = disc_domain_size

        # If label numbers agree with disc_domain_size, then use them.
        # Otherwise, ignore the labels.
        n_env, n_sys = disc_domain_size

        self._env_labels = self._check_labels(n_env, env_labels)
        self._disc_sys_labels = self._check_labels(n_sys, disc_sys_labels)

        # Check each dynamics key is a valid mode,
        # i.e., a valid combination of env and sys labels.
        if dynamics is not None:
            modes = self.all_mode_combs

            undefined_modes = set(dynamics.keys()).difference(modes)

            if undefined_modes:
                msg = 'SwitchedSysDyn: `dynamics` keys inconsistent'
                msg += ' with discrete mode labels.\n'
                msg += 'Undefined modes:\n' + str(undefined_modes)
                raise ValueError(msg)

            missing_modes = set(modes).difference(dynamics.keys())

            if missing_modes:
                msg = 'Missing the modes:\n' + str(missing_modes)
                msg += '\n Make sure you did not forget any modes,\n'
                msg += 'otherwise this is fine.'
                warn(msg)

            if not all([isinstance(sys, PwaSysDyn)
                        for sys in dynamics.values()]):
                msg = 'For each mode dynamics must be PwaSysDyn.\n'
                msg += 'Got instead: ' +str(type(sys))
                raise Exception(msg)

        self.dynamics = dynamics
        self.cts_ss = cts_ss

        _check_time_data(time_semantics, timestep)
        if overwrite_time:
            _push_time_data(self.dynamics.values(), time_semantics, timestep)
        else:
            _check_time_consistency(dynamics.values(), time_semantics, timestep)
        self.timestep = timestep
        self.time_semantics = time_semantics

    def __str__(self):
        n_env, n_sys = self.disc_domain_size

        s = 'Hybrid System Dynamics\n'
        s += 30 * '-' + '\n'

        s += 'Modes:\n'
        s += 4*' ' + 'Environment (' + str(n_env) + ' modes):\n'
        s += 6*' ' + pformat(self.env_labels, indent=3) + 2*'\n'
        s += 4*' ' + 'System: (' + str(n_sys) + ' modes)\n'
        s += 6*' ' + pformat(self.disc_sys_labels, indent=3) + 2*'\n'

        s += 'Continuous State Space:\n\n'
        s += _indent(str(self.cts_ss), 4) + '\n'

        s += 'Dynamics:\n'
        for mode, pwa in self.dynamics.iteritems():
            s += 4*' ' + 'mode: ' + str(mode) + '\n'
            s += 4*' ' + 'dynamics:\n' + _indent(str(pwa), 8) +'\n\n'
        return s

    def _check_labels(self, n, labels):
        # don't complain for default
        if labels is None:
            return None

        # len exists ?
        try:
            # is len correct ?
            if len(labels) != n:
                msg = 'number of environment labels is inconsistent'
                msg += ' with discrete domain size.\n'
                msg += 'Ignoring given environment labels.\n'
                msg += 'Defaulting to integer labels.'
                warn(msg)

                return None
        except:
            warn('Environment labels of type: ' +
                 type(labels) + 'have no len()')
            return None
        return labels

    @property
    def all_mode_combs(self):
        """Return all possible combinations of modes.
        """
        modes = [(a,b) for a in self.env_labels
                           for b in self.disc_sys_labels]

        logger.debug('Available modes: ' + str(modes) )
        return modes

    @property
    def modes(self):
        if self.dynamics is None:
            warn('No dynamics defined (None).')
            return None
        return self.dynamics.keys()

    @property
    def env_labels(self):
        if self._env_labels is None:
            return range(self.disc_domain_size[0])
        else:
            return self._env_labels

    @property
    def disc_sys_labels(self):
        if self._disc_sys_labels is None:
            return range(self.disc_domain_size[1])
        else:
            return self._disc_sys_labels

    @classmethod
    def from_pwa(cls, list_subsys=[], domain=None):
        pwa_sys = PwaSysDyn(list_subsys,domain)
        return cls((1,1), {(0,0):pwa_sys}, domain)

    @classmethod
    def from_lti(cls, A=[], B=[], E=[], K=[],
                 Uset=None, Wset=None,domain=None):
        pwa_sys = PwaSysDyn.from_lti(A, B, E, K,
                                     Uset, Wset, domain)
        return cls((1,1), {(0,0):pwa_sys}, domain)


def _push_time_data(system_list, time_semantics, timestep):
    """Overwrite the time data in system list. Throws warnings if overwriting
    existing data."""

    for system in system_list:
        if (system.time_semantics != time_semantics) and (system.time_semantics
            is not None):
            warn('Overwriting existing time semantics data.')
        if (system.timestep != timestep) and (system.timestep is not None):
            warn('Overwriting existing timestep data.')
        system.time_semantics = time_semantics
        system.timestep = timestep

        # Overwrite LTI in system if system is a PWA
        if isinstance(system, PwaSysDyn):
            _push_time_data(system.list_subsys, time_semantics, timestep)



def _check_time_data(semantics, timestep):
    """Checks that time semantics and timestep are correctly specified. Raises
    ValueErrors if that's not the case.

    @type semantics: string
    @param timestep: any positive number
    @type timestep: int or float or long

    @rtype: None
    """

    if semantics not in ['sampled', 'discrete', None]:
        raise ValueError('Time semantics must be discrete or ' +
            'sampled (sampled from continuous time system).')

    if ((semantics == 'discrete') and (timestep is not None)):
        raise ValueError('Discrete semantics must not have a timestep')

    if timestep is not None:
        error_string = 'Timestep must be a positive real number or unspecified.'
        if timestep <= 0:
            raise ValueError(error_string)
        if not isinstance(timestep, (int, float, long)):
            raise TypeError(error_string)



def _check_time_consistency(system_list, time_semantics, timestep):
    """Checks that all the dynamical systems in system_list have the same time
    semantics and timestep. Raises ValueError if not the case.

    @type system_list: list of L{LtiSysDyn} or L{PwaSysDyn}
    @rtype: None
    """

    # Check that time semantics for all subsystems match
    for ind in range(len(system_list)-1):

        if system_list[ind].timestep != system_list[ind+1].timestep:
            raise ValueError('Not all timesteps in child systems are the same.')

        if system_list[ind].time_semantics != system_list[ind+1].time_semantics:
            raise ValueError('Not all time semantics are the same.')


    # Check that time semantics for all subsystems match specified system and
    # timestep
    if system_list[0].timestep != timestep:
        raise ValueError('Timestep of subsystems do not match specified ' +
                         'timestep.')

    if system_list[0].time_semantics != time_semantics:
        raise ValueError('Time semantics of subsystems do not match ' +
                         'specified time semantics.')



#!/usr/bin/env python
"""
Tests for the interface with Lily.
"""
from nose.tools import raises
from tulip.spec.form import LTL, GRSpec
from tulip.interfaces import lily


def check_realizable(f, expected):
    M = lily.synthesize(f)
    if expected:
        assert M is not None
    else:
        assert M is None

def test_realizable():
    for f, expected in [(LTL('([]<>a) -> ([]<>(a && b))',
                             input_variables={'a': 'boolean'},
                             output_variables={'b': 'boolean'}),
                         True),
                        (LTL('[]<>(a && b)',
                             input_variables={'a': 'boolean'},
                             output_variables={'b': 'boolean'}),
                         False)]:
        yield check_realizable, f, expected

def test_GRSpec():
    f = GRSpec(env_vars={'a'}, sys_vars={'b'},
               env_init=['!a'], sys_init=['!b'],
               env_safety=['a -> X !a', '!a -> X a'],
               sys_prog=['a && b'])
    M = lily.synthesize(f)
    assert M is not None

@raises(TypeError)
def test_nonbool():
    f = LTL('[](a -> <>b)',
            input_variables={'a': (0,1)},
            output_variables={'b': 'boolean'})
    M = lily.synthesize(f)



# Copyright (c) 2011-2014 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""
Functions for plotting Partitions.
"""
import logging
logger = logging.getLogger(__name__)

import numpy as np
from scipy import sparse as sp
import networkx as nx
from polytope.plot import plot_partition, plot_transition_arrow

# inline imports:
#
# import matplotlib as mpl
# from tulip.graphics import newax

def plot_abstraction_scc(ab, ax=None):
    """Plot Regions colored by strongly connected component.

    Handy to develop new examples or debug existing ones.
    """
    try:
        import matplotlib as mpl
    except:
        logger.error('failed to load matplotlib')
        return

    ppp = ab.ppp
    ts = ab.ts
    ppp2ts = ab.ppp2ts

    # each connected component of filtered graph is a symbol
    components = nx.strongly_connected_components(ts)

    if ax is None:
        ax = mpl.pyplot.subplot()

    l, u = ab.ppp.domain.bounding_box
    ax.set_xlim(l[0,0], u[0,0])
    ax.set_ylim(l[1,0], u[1,0])

    for component in components:
        # map to random colors
        red = np.random.rand()
        green = np.random.rand()
        blue = np.random.rand()

        color = (red, green, blue)

        for state in component:
            i = ppp2ts.index(state)
            ppp[i].plot(ax=ax, color=color)
    return ax

def plot_ts_on_partition(ppp, ts, ppp2ts, edge_label, only_adjacent, ax):
    """Plot partition and arrows from labeled digraph.

    Edges can be filtered by selecting an edge_label.
    So it can plot transitions of a single mode for a switched system.

    @param edge_label: desired label
    @type edge_label: dict
    """
    l,u = ppp.domain.bounding_box
    arr_size = (u[0,0]-l[0,0])/50.0

    ts2ppp = {v:k for k,v in enumerate(ppp2ts)}
    for from_state, to_state, label in ts.transitions.find(with_attr_dict=edge_label):
        i = ts2ppp[from_state]
        j = ts2ppp[to_state]

        if only_adjacent:
            if ppp.adj[i, j] == 0:
                continue

        plot_transition_arrow(ppp.regions[i], ppp.regions[j], ax, arr_size)

def project_strategy_on_partition(ppp, mealy):
    """Return an FTS with the PPP (spatial) transitions used by Mealy strategy.

    @type ppp: L{PropPreservingPartition}

    @type mealy: L{transys.MealyMachine}
    """
    n = len(ppp)
    proj_adj = sp.lil_matrix((n, n))

    for (from_state, to_state, label) in mealy.transitions.find():
        from_label = mealy.states[from_state]
        to_label = mealy.states[to_state]

        if 'loc' not in from_label or 'loc' not in to_label:
            continue

        from_loc = from_label['loc']
        to_loc = to_label['loc']

        proj_adj[from_loc, to_loc] = 1

    return proj_adj

def plot_strategy(ab, mealy):
    """Plot strategic transitions on PPP.

    Assumes that mealy is feasible for ab.

    @type ab: L{AbstractPwa} or L{AbstractSwitched}

    @type mealy: L{transys.MealyMachine}
    """
    proj_mealy = project_strategy_on_partition(ab.ppp, mealy)
    ax = plot_partition(ab.ppp, proj_mealy, color_seed=0)
    return ax

def plot_trajectory(ppp, x0, u_seq, ssys,
                    ax=None, color_seed=None):
    """Plots a PropPreservingPartition and the trajectory generated by x0
    input sequence u_seq.

    See Also
    ========
    C{plot_partition}, plot

    @type ppp: L{PropPreservingPartition}
    @param x0: initial state
    @param u_seq: matrix where each row contains an input
    @param ssys: system dynamics
    @param color_seed: see C{plot_partition}
    @return: axis object
    """
    try:
        from tulip.graphics import newax
    except:
        logger.error('failed to import graphics.newax')
        return

    if ax is None:
        ax, fig = newax()

    plot_partition(plot_numbers=False, ax=ax, show=False)

    A = ssys.A
    B = ssys.B

    if ssys.K is not None:
        K = ssys.K
    else:
        K = np.zeros(x0.shape)

    x = x0.flatten()
    x_arr = x0
    for i in range(u_seq.shape[0]):
        x = np.dot(A, x).flatten() +\
            np.dot(B, u_seq[i, :] ).flatten() +\
            K.flatten()
        x_arr = np.vstack([x_arr, x.flatten()])

    ax.plot(x_arr[:,0], x_arr[:,1], 'o-')

    return ax



# Copyright (c) 2013 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Convert state graphs to promela
"""
from time import strftime

def fts2promela(graph, procname=None):
    """Convert (possibly labeled) state graph to Promela str.

    Creates a process which can be simulated as an independent
    thread in the SPIN model checker.
    If atomic propositions label states,
    then they are exported as bit variables.

    The state graph is exported to Promela using goto statements.
    Goto statements avoid introducing additional states.
    Using structured constructs (if, do) can create
    extra states in certain cases.

    The structured program theorem ensures that avoiding goto
    statements is always possible.
    But this yields an equivalent algorithm only,
    not an equivalent unfolding of the program graph.

    Therefore verified properties can be affected,
    especially if intermediate states are introduced in
    one of multiple processes.

    @param graph: networkx

    @param procname: Promela process name (after proctype)
    @type procname: str (default: system name)
    """
    def state_ap2promela(state, graph, ap_alphabet):
        ap_label = get_label_of(state, graph)
        s = ''
        for prop in ap_alphabet:
            if prop is True:
                continue

            if prop in ap_label:
                s += '\t\t ' +str(prop) +' = 1;\n'
            else:
                s += '\t\t ' +str(prop) +' = 0;\n'

        s += '\t\t printf("State: ' +str(state) +'\\n");\n'
        s += '\t\n'
        return s

    def trans2promela(transitions, graph, ap_alphabet):
        s = '\t if\n'
        for (from_state, to_state, sublabels_dict) in transitions:
            s += '\t :: atomic{\n'
            s += '\t\t printf("' +str(sublabels_dict) +'\\n");\n'
            s += state_ap2promela(to_state, graph, ap_alphabet)
            s += '\t\t goto ' +str(to_state) +'\n'
            s += '\t }\n'
        s += '\t fi;\n\n'
        return s

    def get_label_of(state, graph):
        state_label_pairs = graph.states.find([state] )
        (state_, ap_label) = state_label_pairs[0]
        print('state:\t' +str(state) )
        print('ap label:\t' +str(ap_label) )
        return ap_label['ap']

    if procname is None:
        procname = graph.name

    s = '/*\n * Promela file generated with TuLiP\n'
    s += ' * Date: '+str(strftime('%x %X %z') ) +'\n */\n\n'
    for ap in graph.atomic_propositions:
        # convention "!" means negation
        if ap not in {None, True}:
            s += 'bool ' +str(ap) +';\n'

    s += '\nactive proctype ' +procname +'(){\n'

    s += '\t if\n'
    for initial_state in graph.states.initial:
        s += '\t :: goto ' +str(initial_state) +'\n'
    s += '\t fi;\n'

    ap_alphabet = graph.atomic_propositions
    for state in graph.states():
        out_transitions = graph.transitions.find(
            {state}, as_dict=True
        )

        s += str(state).replace(' ', '_') +':'
        s += trans2promela(out_transitions, graph,
                           ap_alphabet)

    s += '}\n'
    return s

#def mealy2promela():
#    """Convert Mealy machine to Promela str.
#    """



# Copyright (c) 2014, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Minimal bindings to external entities."""
from __future__ import absolute_import
from tulip.interfaces import gr1c as gr1cint
from tulip.interfaces import jtlv as jtlvint
from tulip.interfaces._inspect import print_env



# Copyright (c) 2014, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""PLY-based parser for TuLiP LTL syntax,
using AST classes from spec.ast
"""
from __future__ import absolute_import
import logging
logger = logging.getLogger(__name__)
import warnings
import ply.lex
import ply.yacc
import tulip.spec.ast


TABMODULE = 'tulip.spec.ltl_parsetab'
LEX_LOGGER = 'tulip.ltl_lex_log'
YACC_LOGGER = 'tulip.ltl_yacc_log'
PARSER_LOGGER = 'tulip.ltl_parser_log'
# TODO: add past fragment of LTL


class Lexer(object):
    """Token rules to build LTL lexer."""

    reserved = {
        'ite': 'ITE',
        'next': 'NEXT',
        'X': 'NEXT',
        'false': 'FALSE',
        'true': 'TRUE',
        'G': 'ALWAYS',
        'F': 'EVENTUALLY',
        'U': 'UNTIL',
        'W': 'WEAK_UNTIL',
        'V': 'RELEASE'}
    values = {'next': 'X'}
    delimiters = ['LPAREN', 'RPAREN', 'DQUOTES', 'COMMA']
    operators = [
        'NOT', 'AND', 'OR', 'XOR', 'IMP', 'BIMP',
        'EQUALS', 'NEQUALS', 'LT', 'LE', 'GT', 'GE',
        'PLUS', 'MINUS', 'TIMES', 'DIV', 'TRUNCATE', 'PRIME']
    misc = ['NAME', 'NUMBER']

    def __init__(self, debug=False):
        # for setting the logger, call build explicitly
        self.tokens = (
            self.delimiters + self.operators +
            self.misc + list(set(self.reserved.values())))
        self.build(debug=debug)

    def t_NAME(self, t):
        r'[A-Za-z_][A-za-z0-9._:]*'
        t.value = self.values.get(t.value, t.value)
        t.type = self.reserved.get(t.value, 'NAME')
        # special treatment
        if t.value.lower() in {'false', 'true'}:
            t.type = self.reserved[t.value.lower()]
        return t

    # t_PRIME  = r'\''

    def t_ALWAYS(self, t):
        r'\[\]'
        # use single letter as more readable and efficient
        t.value = 'G'
        return t

    def t_EVENTUALLY(self, t):
        r'\<\>'
        t.value = 'F'
        return t

    def t_AND(self, t):
        r'\&\&|\&'
        t.value = '&'
        return t

    def t_OR(self, t):
        r'\|\||\|'
        t.value = '|'
        return t

    t_NOT = r'\!'

    t_XOR = r'\^'

    t_EQUALS = r'\='  # a declarative language has no assignment
    t_NEQUALS = r'\!\='
    t_LT = r'\<'
    t_LE = r'\<\='
    t_GT = r'>\='
    t_GE = r'>'

    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_NUMBER = r'\d+'

    t_IMP = '->'
    t_BIMP = '\<->'

    t_PLUS = r'\+'
    t_MINUS = r'-'
    t_TIMES = r'\*'
    t_DIV = r'/'
    t_TRUNCATE = r'<<>>'
    t_COMMA = r','

    t_DQUOTES = r'\"'
    t_PRIME = r"\'"

    t_ignore = " \t"

    def t_comment(self, t):
        r'\#.*'
        return

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count("\n")

    def t_error(self, t):
        warnings.warn('Illegal character "{t}"'.format(t=t.value[0]))
        t.lexer.skip(1)

    def build(self, debug=False, debuglog=None, **kwargs):
        """Create a lexer.

        @param kwargs: Same arguments as C{ply.lex.lex}:

          - except for C{module} (fixed to C{self})
          - C{debuglog} defaults to the logger C{"ltl_lex_log"}.
        """
        if debug and debuglog is None:
            debuglog = logging.getLogger(LEX_LOGGER)
        self.lexer = ply.lex.lex(
            module=self,
            debug=debug,
            debuglog=debuglog,
            **kwargs)


class Parser(object):
    """Production rules to build LTL parser."""

    tabmodule = TABMODULE
    start = 'expr'
    # lowest to highest
    # closely follows `spin.y`
    precedence = (
        ('left', 'BIMP'),
        ('left', 'IMP'),
        ('left', 'XOR'),
        ('left', 'OR'),
        ('left', 'AND'),
        ('left', 'ALWAYS', 'EVENTUALLY'),
        ('left', 'UNTIL', 'WEAK_UNTIL', 'RELEASE'),
        ('left', 'EQUALS', 'NEQUALS'),
        ('left', 'LT', 'LE', 'GT', 'GE'),
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIV'),
        ('right', 'NOT', 'UMINUS'),
        ('right', 'NEXT'),
        ('left', 'PRIME'),
        ('nonassoc', 'TRUE', 'FALSE'))

    def __init__(self, ast=None, lexer=None):
        if ast is None:
            ast = tulip.spec.ast.nodes
        if lexer is None:
            lexer = Lexer()
        self.ast = ast
        self.lexer = lexer
        self.tokens = self.lexer.tokens
        self.build()

    def build(self, tabmodule=None, outputdir='', write_tables=False,
              debug=False, debuglog=None):
        """Build parser using `ply.yacc`.

        Default table module is `self.tabmodule`.
        Default logger is `YACC_LOGGER`
        """
        if tabmodule is None:
            tabmodule = self.tabmodule
        if debug and debuglog is None:
            debuglog = logging.getLogger(YACC_LOGGER)
        self.parser = ply.yacc.yacc(
            method='LALR',
            module=self,
            start=self.start,
            tabmodule=tabmodule,
            outputdir=outputdir,
            write_tables=write_tables,
            debug=debug,
            debuglog=debuglog)

    def parse(self, formula, debuglog=None):
        """Parse formula string and create abstract syntax tree (AST).

        @param logger: defaults to logger C{"ltl_parser_log"}.
        @type logger: C{logging.Logger}
        """
        if debuglog is None:
            debuglog = logging.getLogger(PARSER_LOGGER)
        root = self.parser.parse(
            formula,
            lexer=self.lexer.lexer,
            debug=debuglog)
        if root is None:
            raise Exception('failed to parse:\n\t{f}'.format(f=formula))
        return root

    def p_nullary_connective(self, p):
        """expr : TRUE
                | FALSE
        """
        p[0] = self.ast.Bool(p[1])

    def p_unary_connective(self, p):
        """expr : NOT expr
                | ALWAYS expr
                | EVENTUALLY expr
                | NEXT expr
        """
        p[0] = self.ast.Unary(p[1], p[2])

    # both function and connective
    def p_postfix_next(self, p):
        """expr : expr PRIME"""
        p[0] = self.ast.Unary('X', p[1])

    def p_binary_connective(self, p):
        """expr : expr AND expr
                | expr OR expr
                | expr XOR expr
                | expr IMP expr
                | expr BIMP expr
                | expr UNTIL expr
                | expr WEAK_UNTIL expr
                | expr RELEASE expr
        """
        p[0] = self.ast.Binary(p[2], p[1], p[3])

    # both function and connective
    def p_ternary_conditional(self, p):
        """expr : LPAREN ITE expr COMMA expr COMMA expr RPAREN"""
        p[0] = self.ast.Operator(p[2], p[3], p[5], p[7])

    def p_binary_predicate(self, p):
        """expr : expr EQUALS expr
                | expr NEQUALS expr
                | expr LT expr
                | expr LE expr
                | expr GT expr
                | expr GE expr
        """
        p[0] = self.ast.Comparator(p[2], p[1], p[3])

    def p_truncator(self, p):
        """expr : expr TRUNCATE number"""
        p[0] = self.ast.Arithmetic(p[2], p[1], p[3])

    def p_binary_function(self, p):
        """expr : expr TIMES expr
                | expr DIV expr
                | expr PLUS expr
                | expr MINUS expr
        """
        p[0] = self.ast.Arithmetic(p[2], p[1], p[3])

    def p_paren(self, p):
        """expr : LPAREN expr RPAREN"""
        p[0] = p[2]

    def p_var(self, p):
        """expr : NAME"""
        p[0] = self.ast.Var(p[1])

    def p_number_expr(self, p):
        """expr : number"""
        p[0] = p[1]

    def p_number(self, p):
        """number : NUMBER"""
        p[0] = self.ast.Num(p[1])

    def p_negative_number(self, p):
        """expr : MINUS NUMBER %prec UMINUS"""
        p[0] = self.ast.Num('-' + p[2])

    def p_string(self, p):
        """expr : DQUOTES NAME DQUOTES"""
        p[0] = self.ast.Str(p[2])

    def p_error(self, p):
        s = list()
        while True:
            tok = self.parser.token()
            if tok is None:
                break
            s.append(tok.value)
        raise Exception(
            'Syntax error at "{p}"\n'.format(p=p.value) +
            'remaining input:\n{s}\n'.format(s=' '.join(s)))


def parse(formula):
    warnings.warn('Deprecated: Better to instantiate a Parser once only.')
    parser = Parser()
    return parser.parse(formula)


if __name__ == '__main__':
    h = logging.FileHandler('log.txt', mode='w')
    h.setLevel(logging.DEBUG)
    log = logging.getLogger(YACC_LOGGER)
    log.setLevel(logging.DEBUG)
    log.addHandler(h)
    import os
    tabmodule = TABMODULE.split('.')[-1]
    outputdir = './'
    tablepy = tabmodule + '.py'
    tablepyc = tabmodule + '.pyc'
    try:
        os.remove(tablepy)
    except:
        print('no "{t}" found'.format(t=tablepy))
    try:
        os.remove(tablepyc)
    except:
        print('no "{t}" found'.format(t=tablepyc))
    parser = Parser()
    parser.build(tabmodule, outputdir=outputdir,
                 write_tables=True, debug=True)



# Copyright (c) 2011-2014 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""LTL parser supporting JTLV, SPIN, SMV, and gr1c syntax"""
from __future__ import absolute_import
import re
from tulip.spec import ast, lexyacc


# cache
parsers = dict()


def parse(formula, full_operators=False):
    """Parse formula string and create abstract syntax tree (AST).

    @param full_operators: replace full names of operators
        with their symbols (case insensitive,
        each operator must be a separate word).
    @type full_operators: C{bool}
    """
    if full_operators:
        formula = _replace_full_name_operators(formula)
    if parsers.get('ply') is None:
        parsers['ply'] = lexyacc.Parser()
    spec = parsers['ply'].parse(formula)
    # did ply fail merely printing warnings ?
    if spec is None:
        raise Exception('Parsing formula:\n{f}\nfailed'.format(f=formula))
    return spec


def _replace_full_name_operators(formula):
    """Replace full names with symbols for temporal and Boolean operators.

    Each operator must be a word (as defined by \b in regexp).
    Substitution is case insensitive.
    """
    for name, symbol in ast.FULL_OPERATOR_NAMES.iteritems():
        formula = re.sub(r'\b(?i)' + name + r'\b', symbol, formula)
    return formula



#!/usr/bin/env python
"""
Tests for the interface with JTLV.
"""

import networkx as nx
import nose.tools as nt
from tulip.spec import GRSpec
import tulip.interfaces.jtlv as jtlv
import networkx as nx


class basic_test(object):
    def setUp(self):
        self.check_realizable = jtlv.check_realizable
        self.synthesize = jtlv.synthesize
        self.f_un = GRSpec(env_vars="x", sys_vars="y",
                           env_init="x", env_prog="x",
                           sys_init="y", sys_safety=["y -> X(!y)", "!y -> X(y)"],
                           sys_prog="y && x",
                           moore=False, plus_one=False, qinit='\A \E')
        self.f = GRSpec(env_vars="x", sys_vars="y",
                        env_init="x", env_prog="x",
                        sys_init="y",
                        sys_prog=["y & x", "!y"],
                        moore=False, plus_one=False, )
        self.dcounter = GRSpec(sys_vars={"y": (0,5)}, sys_init=["y=0"],
                               sys_prog=["y=0", "y=5"],
                               moore=False, plus_one=False, qinit='\A \E')

    def tearDown(self):
        self.f_un = None
        self.f = None
        self.dcounter = None

    def test_check_realizable(self):
        assert not self.check_realizable(self.f_un)
        self.f_un.sys_safety = []
        assert self.check_realizable(self.f_un)
        assert self.check_realizable(self.dcounter)

    def test_synthesize(self):
        g = self.synthesize(self.f_un)
        assert not isinstance(g, nx.DiGraph)

        g = self.synthesize(self.f)
        # There is more than one possible strategy realizing this
        # specification.  Checking only for one here makes this more like
        # a regression test (fragile).  However, it is more meaningful
        # than simply checking that synthesize() returns something
        # non-None (i.e., realizability, which is tested elsewhere).
        assert g is not None

        # assert len(g.env_vars) == 1 and g.env_vars.has_key('x')
        # assert len(g.sys_vars) == 1 and g.sys_vars.has_key('y')
        print(g.nodes())
        assert len(g) == 5
        assert set(g.edges()) == set([(0, 1), (0, 2), (1, 3), (1, 4),
                                              (2, 3), (2, 4), (3, 0), (3, 3),
                                              (4, 0), (4, 3)])
        label_reference = {0: (1,1), # value is bitvector of x,y
                           1: (0,0),
                           2: (1,0),
                           3: (0,0),
                           4: (1,0)}
        for u, d in g.nodes_iter(data=True):
            state = d['state']
            assert(len(state) == 2)
            assert(label_reference[u] == (state['x'], state['y']))


def hash_question_mark_test():
    specs = GRSpec(env_vars={'w': ['low', 'medium', 'high']},
                   sys_vars={'a': (0, 2)},
                   # env
                   env_init=['w="low"'],
                   env_safety=['(a=1) -> ((w="low") || (w="medium"))'],
                   env_prog=['(w="high")'],
                   sys_init=['a=2'],
                   sys_safety=['a=2'],
                   sys_prog=['a=2'],
                   moore=False, plus_one=False, qinit='\A \E')
    with nt.assert_raises(ValueError):
        jtlv.synthesize(specs)



# Copyright (c) 2013-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Algorithms on Kripke structures and Automata"""
from __future__ import absolute_import
import logging
import copy
from tulip.transys.transys import FiniteTransitionSystem
from tulip.transys.automata import BuchiAutomaton
import networkx as nx
from tulip.interfaces import ltl2ba as ltl2baint
# possible future:
# from tulip.transys.transys import TransitionSystem
# from tulip.transys.automata import Automaton
from tulip.transys.labeled_graphs import LabeledDiGraph


_hl = 40 * '-'
logger = logging.getLogger(__name__)


# build parser once only
parser = ltl2baint.Parser()


def ltl2ba(formula):
    """Convert LTL formula to Buchi Automaton using ltl2ba.

    @type formula: `str(formula)` must be admissible ltl2ba input

    @return: Buchi automaton whose edges are annotated
        with Boolean formulas as `str`
    @rtype: [`Automaton`]
    """
    ltl2ba_out = ltl2baint.call_ltl2ba(str(formula))
    symbols, g, initial, accepting = parser.parse(ltl2ba_out)
    ba = Automaton('Buchi', alphabet=symbols)
    ba.add_nodes_from(g)
    ba.add_edges_from(g.edges_iter(data=True))
    ba.initial_nodes = initial
    ba.accepting_sets = accepting
    logger.info('Resulting automaton:\n\n{ba}\n'.format(ba=ba))
    return ba


def _multiply_mutable_states(self, other, prod_graph, prod_sys):
    def prod_ids2states(prod_state_id, self, other):
        (idx1, idx2) = prod_state_id
        state1 = self.states._int2mutant(idx1)
        state2 = other.states._int2mutant(idx2)
        prod_state = (state1, state2)
        return prod_state

    def label_union(nx_label):
        (v1, v2) = nx_label

        if v1 is None or v2 is None:
            raise Exception(
                'At least one factor has unlabeled state, '
                "or the state sublabel types don't match.")
        try:
            return v1 | v2
        except:
            pass
        try:
            return v2 + v2
        except:
            raise TypeError(
                'The state sublabel types should support ' +
                'either | or + for labeled system products.')

    def state_label_union(attr_dict):
        prod_attr_dict = dict()
        for k, v in attr_dict.iteritems():
            prod_attr_dict[k] = label_union(v)
        return prod_attr_dict

    # union of state labels from the networkx tuples
    for prod_state_id, attr_dict in prod_graph.nodes_iter(data=True):
        prod_attr_dict = state_label_union(attr_dict)
        prod_state = prod_ids2states(prod_state_id, self, other)

        prod_sys.states.add(prod_state)
        prod_sys.states.add(prod_state, **prod_attr_dict)
    print(prod_sys.states)

    # prod of initial states
    inits1 = self.states.initial
    inits2 = other.states.initial

    prod_init = []
    for (init1, init2) in zip(inits1, inits2):
        new_init = (init1, init2)
        prod_init.append(new_init)
    prod_sys.states.initial |= prod_init

    # # multiply mutable states (only the reachable added)
    # if self.states.mutants or other.states.mutants:
    #     for idx, prod_state_id in enumerate(prod_graph.nodes_iter() ):
    #         prod_state = prod_ids2states(prod_state_id, self, other)
    #         prod_sys.states.mutants[idx] = prod_state
    #
    #     prod_sys.states.min_free_id = idx +1
    # # no else needed: otherwise self already not mutant

    # action labeling is taken care by nx,
    # since transition taken at a time
    for from_state_id, to_state_id, edge_dict in prod_graph.edges_iter(data=True):

        from_state = prod_ids2states(from_state_id, self, other)
        to_state = prod_ids2states(to_state_id, self, other)

        prod_sys.transitions.add(
            from_state, to_state, **edge_dict)
    return prod_sys


# binary operators (for magic binary operators: see above)
def tensor_product(self, other, prod_sys=None):
    """Return strong product with given graph.

    Reference
    =========
    http://en.wikipedia.org/wiki/Strong_product_of_graphs
    nx.algorithms.operators.product.strong_product
    """
    prod_graph = nx.product.tensor_product(self, other)
    # not populating ?
    if prod_sys is None:
        if self.states.mutants or other.states.mutants:
            mutable = True
        else:
            mutable = False
        prod_sys = LabeledDiGraph(mutable=mutable)
    prod_sys = self._multiply_mutable_states(
        other, prod_graph, prod_sys)
    return prod_sys


def cartesian_product(self, other, prod_sys=None):
    """Return Cartesian product with given graph.

    If u,v are nodes in C{self} and z,w nodes in C{other},
    then ((u,v), (z,w) ) is an edge in the Cartesian product of
    self with other if and only if:
        - (u == v) and (z,w) is an edge of C{other}
        OR
        - (u,v) is an edge in C{self} and (z == w)

    In system-theoretic terms, the Cartesian product
    is the interleaving where at each step,
    only one system/process/player makes a move/executes.

    So it is a type of parallel system.

    This is an important distinction with the C{strong_product},
    because that includes "diagonal" transitions, i.e., two
    processes executing truly concurrently.

    Note that a Cartesian interleaving is different from a
    strong interleaving, because the latter can skip states
    and transition directly along the diagonal.

    For a model of computation, strong interleaving
    would accurately model the existence of multiple cores,
    not just multiple processes executing on a single core.

    References
    ==========
      - U{http://en.wikipedia.org/wiki/Cartesian_product_of_graphs}
      - networkx.algorithms.operators.product.cartesian_product
    """
    prod_graph = nx.product.cartesian_product(self, other)
    # not populating ?
    if prod_sys is None:
        if self.states.mutants or other.states.mutants:
            mutable = True
        else:
            mutable = False
        prod_sys = LabeledDiGraph(mutable=mutable)
    prod_sys = self._multiply_mutable_states(
        other, prod_graph, prod_sys)
    return prod_sys


def ts_sync_prod(ts1, ts2):
    """Synchronous (tensor) product with other FTS.

    @type ts1, ts2: L{FiniteTransitionSystem}
    """
    prod_ts = FiniteTransitionSystem()
    # union of AP sets
    prod_ts.atomic_propositions |= \
        ts1.atomic_propositions | ts2.atomic_propositions
    # use more label sets, instead of this explicit approach
    #
    # for synchronous product: Cartesian product of action sets
    # prod_ts.actions |= ts1.actions * ts2.actions
    prod_ts = super(FiniteTransitionSystem, self).tensor_product(
        ts, prod_sys=prod_ts)
    return prod_ts


def sync_prod(ts, ba):
    """Synchronous product between (BA, TS), or (BA1, BA2).

    The result is always a L{BuchiAutomaton}:

        - If C{ts_or_ba} is a L{FiniteTransitionSystem} TS,
            then return the synchronous product BA * TS.

            The accepting states of BA * TS are those which
            project on accepting states of BA.

        - If C{ts_or_ba} is a L{BuchiAutomaton} BA2,
            then return the synchronous product BA * BA2.

            The accepting states of BA * BA2 are those which
            project on accepting states of both BA and BA2.

            This definition of accepting set extends
            Def.4.8, p.156 U{[BK08]
            <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
            to NBA.

    Synchronous product TS * BA or TS1 * TS2.

    Returns a Finite Transition System, because TS is
    the first term in the product.

    Changing term order, i.e., BA * TS, returns the
    synchronous product as a BA.

    Caution
    =======
    This method includes semantics for true\in\Sigma (p.916, U{[BK08]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}),
    so there is a slight overlap with logic grammar.  In other
    words, not completely isolated from logics.

    See Also
    ========
    L{_ts_ba_sync_prod}

    @param ts_or_ba: other with which to take synchronous product
    @type ts_or_ba: L{FiniteTransitionSystem} or L{BuchiAutomaton}

    @return: self * ts_or_ba
    @rtype: L{BuchiAutomaton}

    See Also
    ========
    __mul__, async_prod, BuchiAutomaton.sync_prod, tensor_product
    Def. 2.42, pp. 75--76 U{[BK08]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
    Def. 4.62, p.200 U{[BK08]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}

    @param ts_or_ba: system with which to take synchronous product
    @type ts_or_ba: L{FiniteTransitionSystem} or L{BuchiAutomaton}

    @return: synchronous product C{self} x C{ts_or_ba}
    @rtype: L{FiniteTransitionSystem}
    """
    if not isinstance(ba, BuchiAutomaton):
        raise Exception
    if not isinstance(ts, FiniteTransitionSystem):
        raise Exception


def add(self, other):
    """Merge two Finite Transition Systems.

    States, Initial States, Actions, Atomic Propositions and
    State labels and Transitions of the second Transition System
    are merged into the first and take precedence, overwriting
    existing labeling.

    Example
    =======
    This can be useful to construct systems quickly by creating
    standard "pieces" using the functions: line_labeled_with,
    cycle_labeled_with

    >>> n = 4
    >>> L = n*['p'] # state labeling
    >>> ts1 = line_labeled_with(L, n-1)
    >>> ts1.plot()
    >>>
    >>> L = n*['p']
    >>> ts2 = cycle_labeled_with(L)
    >>> ts2.states.add('s3', ap={'!p'})
    >>> ts2.plot()
    >>>
    >>> ts3 = ts1 +ts2
    >>> ts3.transitions.add('s'+str(n-1), 's'+str(n) )
    >>> ts3.default_layout = 'circo'
    >>> ts3.plot()

    See Also
    ========
    L{line_labeled_with}, L{cycle_labeled_with}

    @param other: system to merge with
    @type other: C{FiniteTransitionSystem}

    @return: merge of C{self} with C{other}, union of states,
        initial states, atomic propositions, actions, edges and
        labelings, those of C{other} taking precedence over C{self}.
    @rtype: L{FiniteTransitionSystem}
    """
    if not isinstance(other, FiniteTransitionSystem):
        msg = (
            'other class must be FiniteTransitionSystem.\n'
            'Got instead:\n\t' + str(other) +
            '\nof type:\n\t' + str(type(other)))
        raise TypeError(msg)
    self.atomic_propositions |= other.atomic_propositions
    self.actions |= other.actions
    # add extra states & their labels
    for state, label in other.states.find():
        if state not in self.states:
            self.states.add(state)
        if label:
            self.states[state]['ap'] = label['ap']
    self.states.initial |= set(other.states.initial)
    # copy extra transitions (be careful w/ labeling)
    for (from_state, to_state, label_dict) in other.transitions.find():
        # labeled edge ?
        if not label_dict:
            self.transitions.add(from_state, to_state)
        else:
            sublabel_value = label_dict['actions']
            self.transitions.add(
                from_state, to_state, actions=sublabel_value)
    return copy.copy(self)


def async_prod(self, ts):
    """Asynchronous product TS1 x TS2 between FT Systems.

    See Also
    ========
    __or__, sync_prod, cartesian_product
    Def. 2.18, p.38 U{[BK08]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#bk08>}
    """
    if not isinstance(ts, FiniteTransitionSystem):
        raise TypeError('ts must be a FiniteTransitionSystem.')
    if self.states.mutants or ts.states.mutants:
        mutable = True
    else:
        mutable = False
    # union of AP sets
    prod_ts = FiniteTransitionSystem(mutable=mutable)
    prod_ts.atomic_propositions |= \
        self.atomic_propositions | ts.atomic_propositions
    # for parallel product: union of action sets
    prod_ts.actions |= self.actions | ts.actions
    prod_ts = super(FiniteTransitionSystem, self).cartesian_product(
        ts, prod_sys=prod_ts)
    return prod_ts



#!/usr/bin/env python
"""Example using transition system dynamics.

This example illustrates the use of TuLiP to synthesize a reactive
controller for system whose dynamics are described by a discrete
transition system.
"""
# RMM, 20 Jul 2013
#
# Note: This code is commented to allow components to be extracted into
# the tutorial that is part of the users manual.  Comments containing
# strings of the form @label@ are used for this purpose.

# @import_section@
# Import the packages that we need
import logging

from tulip import transys, spec, synth
# @import_section_end@


logging.basicConfig(level=logging.WARNING)
logging.getLogger('tulip.spec.lexyacc').setLevel(logging.WARNING)
logging.getLogger('tulip.synth').setLevel(logging.WARNING)
logging.getLogger('tulip.interfaces.omega').setLevel(logging.WARNING)


#
# System dynamics
#
# The system is modeled as a discrete transition system in which the
# robot can be located anyplace on a 2x3 grid of cells.  Transitions
# between adjacent cells are allowed, which we model as a transition
# system in this example (it would also be possible to do this via a
# formula)
#
# We label the states using the following picture
#
#     +----+----+----+
#     | X3 | X4 | X5 |
#     +----+----+----+
#     | X0 | X1 | X2 |
#     +----+----+----+
#

# @system_dynamics_section@
# Create a finite transition system
sys = transys.FTS()

# Define the states of the system
sys.states.add_from(['X0', 'X1', 'X2', 'X3', 'X4', 'X5'])
sys.states.initial.add('X0')    # start in state X0

# Define the allowable transitions
#! TODO (IF): can arguments be a singleton instead of a list?
#! TODO (IF): can we use lists instead of sets?
#!   * use optional flag to allow list as label
sys.transitions.add_comb({'X0'}, {'X1', 'X3'})
sys.transitions.add_comb({'X1'}, {'X0', 'X4', 'X2'})
sys.transitions.add_comb({'X2'}, {'X1', 'X5'})
sys.transitions.add_comb({'X3'}, {'X0', 'X4'})
sys.transitions.add_comb({'X4'}, {'X3', 'X1', 'X5'})
sys.transitions.add_comb({'X5'}, {'X4', 'X2'})
# @system_dynamics_section_end@

# @system_labels_section@
# Add atomic propositions to the states
sys.atomic_propositions.add_from({'home', 'lot'})
sys.states.add('X0', ap={'home'})
sys.states.add('X5', ap={'lot'})
# @system_labels_section_end@

# if IPython and Matplotlib available
#sys.plot()

#
# Environment variables and specification
#
# The environment can issue a park signal that the robot must respond
# to by moving to the lower left corner of the grid.  We assume that
# the park signal is turned off infinitely often.
#
# @environ_section@
env_vars = {'park'}
env_init = set()                # empty set
env_prog = '!park'
env_safe = set()                # empty set
# @environ_section_end@

#
# System specification
#
# The system specification is that the robot should repeatedly revisit
# the upper right corner of the grid while at the same time responding
# to the park signal by visiting the lower left corner.  The LTL
# specification is given by
#
#     []<> home && [](park -> <>lot)
#
# Since this specification is not in GR(1) form, we introduce the
# variable X0reach that is initialized to True and the specification
# [](park -> <>lot) becomes
#
#     [](X (X0reach) <-> lot || (X0reach && !park))
#

# @specs_setup_section@
# Augment the system description to make it GR(1)
#! TODO: create a function to convert this type of spec automatically
sys_vars = {'X0reach'}          # infer the rest from TS
sys_init = {'X0reach'}
sys_prog = {'home'}             # []<>home
sys_safe = {'(X (X0reach) <-> lot) || (X0reach && !park)'}
sys_prog |= {'X0reach'}
# @specs_setup_section_end@

# @specs_create_section@
# Create the specification
specs = spec.GRSpec(env_vars, sys_vars, env_init, sys_init,
                    env_safe, sys_safe, env_prog, sys_prog)
# @specs_create_section_end@

#
# Controller synthesis
#
# At this point we can synthesize the controller using one of the available
# methods.
#
# @synthesize@
# Moore machines
# controller reads `env_vars, sys_vars`, but not next `env_vars` values
specs.moore = True
# synthesizer should find initial system values that satisfy
# `env_init /\ sys_init` and work, for every environment variable
# initial values that satisfy `env_init`.
specs.qinit = '\E \A'
ctrl = synth.synthesize('omega', specs, sys=sys)
assert ctrl is not None, 'unrealizable'
# @synthesize_end@

#
# Generate a graphical representation of the controller for viewing,
# or a textual representation if pydot is missing.
#
# @plot_print@
if not ctrl.save('discrete.png'):
    print(ctrl)
# @plot_print_end@



# Copyright (c) 2011-2014 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""
Algorithms related to controller synthesis for discretized dynamics.

Primary functions:
    - L{get_input}

Helper functions:
    - L{get_input_helper}
    - L{is_seq_inside}

See Also
========
L{discretize}
"""
from __future__ import absolute_import
import logging

import numpy as np
import polytope as pc
try:
    from cvxopt import matrix, solvers
except ImportError:
    solvers = None

from tulip.abstract.feasible import (
    solve_feasible,
    createLM,
    _block_diag2)


logger = logging.getLogger(__name__)
if solvers is None:
    logger.warn(
        '`tulip` failed to import `cvxopt`.\n'
        'No quadratic cost for controller computation.')


def assert_cvxopt():
    """Raise `ImportError` if `cvxopt.solvers` failed to import."""
    if solvers is None:
        raise ImportError(
            'Failed to import `cvxopt.solvers`.'
            'Unable to solve quadratic programming problems.')


def get_input(
    x0, ssys, abstraction,
    start, end,
    R=None, r=None, Q=None,
    ord=1, mid_weight=0.0
):
    """Compute continuous control input for discrete transition.

    Computes a continuous control input sequence
    which takes the plant:

        - from state C{start}
        - to state C{end}

    These are states of the partition C{abstraction}.
    The computed control input is such that::

        f(x, u) = |Rx|_{ord} + |Qu|_{ord} + r'x +
                  mid_weight * |xc - x(N)|_{ord}

    be minimal.

    C{xc} is the chebyshev center of the final cell.
    If no cost parameters are given, then the defaults are:

        - Q = I
        - mid_weight = 3

    Notes
    =====
    1. The same horizon length as in reachability analysis
        should be used in order to guarantee feasibility.

    2. If the closed loop algorithm has been used
        to compute reachability the input needs to be
        recalculated for each time step
        (with decreasing horizon length).

        In this case only u(0) should be used as
        a control signal and u(1) ... u(N-1) discarded.

    3. The "conservative" calculation makes sure that
        the plant remains inside the convex hull of the
        starting region during execution, i.e.::

            x(1), x(2) ...  x(N-1) are
            \in conv_hull(starting region).

        If the original proposition preserving partition
        is not convex, then safety cannot be guaranteed.

    @param x0: initial continuous state
    @type x0: numpy 1darray

    @param ssys: system dynamics
    @type ssys: L{LtiSysDyn}

    @param abstraction: abstract system dynamics
    @type abstraction: L{AbstractPwa}

    @param start: index of the initial state in C{abstraction.ts}
    @type start: int >= 0

    @param end: index of the end state in C{abstraction.ts}
    @type end: int >= 0

    @param R: state cost matrix for::
            x = [x(1)' x(2)' .. x(N)']'
        If empty, zero matrix is used.
    @type R: size (N*xdim x N*xdim)

    @param r: cost vector for state trajectory:
        x = [x(1)' x(2)' .. x(N)']'
    @type r: size (N*xdim x 1)

    @param Q: input cost matrix for control input::
            u = [u(0)' u(1)' .. u(N-1)']'
        If empty, identity matrix is used.
    @type Q: size (N*udim x N*udim)

    @param mid_weight: cost weight for |x(N)-xc|_{ord}

    @param ord: norm used for cost function::
        f(x, u) = |Rx|_{ord} + |Qu|_{ord} + r'x +
              mid_weight *|xc - x(N)|_{ord}
    @type ord: ord \in {1, 2, np.inf}

    @return: array A where row k contains the
        control input: u(k)
        for k = 0, 1 ... N-1
    @rtype: (N x m) numpy 2darray
    """
    part = abstraction.ppp
    regions = part.regions

    ofts = abstraction.ts
    original_regions = abstraction.orig_ppp
    orig = abstraction._ppp2orig

    params = abstraction.disc_params
    N = params['N']  # horizon length
    conservative = params['conservative']
    closed_loop = params['closed_loop']
    if closed_loop:
        logger.warn(
            '`closed_loop = True` for controller computation. '
            'This option is under development: use with caution.')
    if (
            R is None and
            Q is None and
            r is None and
            mid_weight == 0):
        # Default behavior
        Q = np.eye(N * ssys.B.shape[1])
        R = np.zeros([N * x0.size, N * x0.size])
        r = np.zeros([N * x0.size, 1])
        mid_weight = 3
    if R is None:
        R = np.zeros([N * x0.size, N * x0.size])
    if Q is None:
        Q = np.eye(N * ssys.B.shape[1])
    if r is None:
        r = np.zeros([N * x0.size, 1])

    if (R.shape[0] != R.shape[1]) or (R.shape[0] != N * x0.size):
        raise Exception("get_input: "
                        "R must be square and have side N * dim(state space)")

    if (Q.shape[0] != Q.shape[1]) or (Q.shape[0] != N * ssys.B.shape[1]):
        raise Exception("get_input: "
                        "Q must be square and have side N * dim(input space)")
    if ofts is not None:
        start_state = start
        end_state = end

        if end_state not in ofts.states.post(start_state):
            raise Exception('get_input: '
                            'no transition from state s' + str(start) +
                            ' to state s' + str(end)
                            )
    else:
        print("get_input: "
              "Warning, no transition matrix found, assuming feasible")

    if (not conservative) & (orig is None):
        print("List of original proposition preserving "
              "partitions not given, reverting to conservative mode")
        conservative = True

    P_start = regions[start]
    P_end = regions[end]

    n = ssys.A.shape[1]
    m = ssys.B.shape[1]

    idx = range((N - 1) * n, N * n)

    if conservative:
        # Take convex hull or P_start as constraint
        if len(P_start) > 0:
            if len(P_start) > 1:
                # Take convex hull
                vert = pc.extreme(P_start[0])
                for i in range(1, len(P_start)):
                    vert = np.vstack([
                        vert,
                        pc.extreme(P_start[i])
                    ])
                P1 = pc.qhull(vert)
            else:
                P1 = P_start[0]
        else:
            P1 = P_start
    else:
        # Take original proposition preserving cell as constraint
        P1 = original_regions[orig[start]]
        # must be single polytope (ensuring convex)
        assert len(P1) > 0, P1
        if len(P1) == 1:
            P1 = P1[0]
        else:
            print(P1)
            raise Exception(
                '`conservative = False` arg requires '
                'that original regions be convex')

    if len(P_end) > 0:
        low_cost = np.inf
        low_u = np.zeros([N, m])

        # for each polytope in target region
        for P3 in P_end:
            if mid_weight > 0:
                rc, xc = pc.cheby_ball(P3)
                R[
                    np.ix_(
                        range(n * (N - 1), n * N),
                        range(n * (N - 1), n * N)
                    )
                ] += mid_weight * np.eye(n)

                r[idx, :] += -mid_weight * xc
                u, cost = get_input_helper(
                    x0, ssys, P1, P3, N, R, r, Q, ord,
                    closed_loop=closed_loop
                )
                r[idx, :] += mid_weight * xc

            if cost < low_cost:
                low_u = u
                low_cost = cost

        if low_cost == np.inf:
            raise Exception("get_input: Did not find any trajectory")
    else:
        P3 = P_end
        if mid_weight > 0:
            rc, xc = pc.cheby_ball(P3)
            R[
                np.ix_(
                    range(n * (N - 1), n * N),
                    range(n * (N - 1), n * N)
                )
            ] += mid_weight * np.eye(n)
            r[idx, :] += -mid_weight * xc
        low_u, cost = get_input_helper(
            x0, ssys, P1, P3, N, R, r, Q, ord,
            closed_loop=closed_loop
        )
    return low_u


def get_input_helper(
    x0, ssys, P1, P3, N, R, r, Q, ord=1,
    closed_loop=True
):
    """Calculate the sequence u_seq such that:

      - x(t+1) = A x(t) + B u(t) + K
      - x(k) \in P1 for k = 0,...N
      - x(N) \in P3
      - [u(k); x(k)] \in PU

    and minimize:
        |Rx|_{ord} + |Qu|_{ord} + r'x +
        mid_weight * |xc - x(N)|_{ord}
    """
    n = ssys.A.shape[1]
    m = ssys.B.shape[1]

    list_P = []
    if closed_loop:
        temp_part = P3
        list_P.append(P3)
        for i in xrange(N - 1, 0, -1):
            temp_part = solve_feasible(
                P1, temp_part, ssys, N=1,
                closed_loop=False, trans_set=P1
            )
            list_P.insert(0, temp_part)
        list_P.insert(0, P1)
        L, M = createLM(ssys, N, list_P, disturbance_ind=[1])
    else:
        list_P.append(P1)
        for i in xrange(N - 1, 0, -1):
            list_P.append(P1)
        list_P.append(P3)
        L, M = createLM(ssys, N, list_P)

    # Remove first constraint on x(0)
    L = L[range(list_P[0].A.shape[0], L.shape[0]), :]
    M = M[range(list_P[0].A.shape[0], M.shape[0]), :]

    # Separate L matrix
    Lx = L[:, range(n)]
    Lu = L[:, range(n, L.shape[1])]
    M = M - Lx.dot(x0).reshape(Lx.shape[0], 1)

    B_diag = ssys.B
    for i in xrange(N - 1):
        B_diag = _block_diag2(B_diag, ssys.B)
    K_hat = np.tile(ssys.K, (N, 1))
    A_it = ssys.A.copy()
    A_row = np.zeros([n, n * N])
    A_K = np.zeros([n * N, n * N])
    A_N = np.zeros([n * N, n])

    for i in xrange(N):
        A_row = ssys.A.dot(A_row)
        A_row[np.ix_(
            range(n),
            range(i * n, (i + 1) * n)
        )] = np.eye(n)

        A_N[np.ix_(
            range(i * n, (i + 1) * n),
            range(n)
        )] = A_it

        A_K[np.ix_(
            range(i * n, (i + 1) * n),
            range(A_K.shape[1])
        )] = A_row

        A_it = ssys.A.dot(A_it)
    Ct = A_K.dot(B_diag)
    if ord == 1:
        # f(\epsilon,u) = sum(\epsilon)
        c_LP = np.hstack((np.ones((1, N * (n + m))), r.T.dot(Ct)))
        # Constraints -\epsilon_r < R*x <= \epsilon_r
        # Constraints -\epsilon_u < Q*u <= \epsilon_u
        # x = A_N*x0 + Ct*u --> ignore the first constant part
        # x  = Ct*u
        G_LP = np.vstack((
            np.hstack((- np.eye(N * n),
                       np.zeros((N * n, N * m)),
                       - R.dot(Ct))),
            np.hstack((- np.eye(N * n),
                       np.zeros((N * n, N * m)),
                       R.dot(Ct))),
            np.hstack((np.zeros((N * m, N * n)),
                       - np.eye(N * m), -Q)),
            np.hstack((np.zeros((N * m, N * n)),
                       - np.eye(N * m), Q)),
            np.hstack((np.zeros((Lu.shape[0], N * n + N * m)), Lu))
        ))
        h_LP = np.vstack((np.zeros((2 * N * (n + m), 1)), M))
    elif ord == 2:
        assert_cvxopt()
        # symmetrize
        Q2 = Q.T.dot(Q)
        R2 = R.T.dot(R)
        # constraints
        G = matrix(Lu)
        h = matrix(M)
        P = matrix(Q2 + Ct.T.dot(R2).dot(Ct))
        q = matrix(
            np.dot(
                np.dot(x0.reshape(1, x0.size), A_N.T) +
                A_K.dot(K_hat).T,
                R2.dot(Ct)
            ) +
            0.5 * r.T.dot(Ct)
        ).T
        sol = solvers.qp(P, q, G, h)
        if sol['status'] != "optimal":
            raise Exception(
                "getInputHelper: "
                "QP solver finished with status " +
                str(sol['status']))
        u = np.array(sol['x']).flatten()
        cost = sol['primal objective']
        return u.reshape(N, m), cost
    elif ord == np.inf:
        c_LP = np.hstack((np.ones((1, 2)), r.T.dot(Ct)))
        G_LP = np.vstack((
            np.hstack((-np.ones((N * n, 1)),
                       np.zeros((N * n, 1)),
                       -R.dot(Ct))),
            np.hstack((-np.ones((N * n, 1)),
                       np.zeros((N * n, 1)),
                       R.dot(Ct))),
            np.hstack((np.zeros((N * m, 1)),
                       -np.ones((N * m, 1)), -Q)),
            np.hstack((np.zeros((N * m, 1)),
                       -np.ones((N * m, 1)), Q)),
            np.hstack((np.zeros((Lu.shape[0], 2)), Lu))
        ))
        h_LP = np.vstack((np.zeros((2 * N * (n + m), 1)), M))
    sol = pc.polytope.lpsolve(c_LP.flatten(), G_LP, h_LP)
    if sol['status'] != 0:
        raise Exception(
            "getInputHelper: "
            "LP solver finished with message " +
            str(sol['message']))
    var = np.array(sol['x']).flatten()
    u = var[-N * m:]
    cost = sol['fun']
    return u.reshape(N, m), cost


def is_seq_inside(x0, u_seq, ssys, P0, P1):
    """Checks if the plant remains inside P0 for time t = 1, ... N-1
    and  that the plant reaches P1 for time t = N.
    Used to test a computed input sequence.
    No disturbance is taken into account.

    @param x0: initial point for execution
    @param u_seq: (N x m) array where row k is input for t = k

    @param ssys: dynamics
    @type ssys: L{LtiSysDyn}

    @param P0: C{Polytope} where we want x(k) to remain for k = 1, ... N-1

    @return: C{True} if x(k) \in P0 for k = 1, .. N-1 and x(N) \in P1.
        C{False} otherwise
    """
    N = u_seq.shape[0]
    x = x0.reshape(x0.size, 1)

    A = ssys.A
    B = ssys.B
    if len(ssys.K) == 0:
        K = np.zeros(x.shape)
    else:
        K = ssys.K

    inside = True
    for i in xrange(N - 1):
        u = u_seq[i, :].reshape(u_seq[i, :].size, 1)
        x = A.dot(x) + B.dot(u) + K

        if not pc.is_inside(P0, x):
            inside = False

    un_1 = u_seq[N - 1, :].reshape(u_seq[N - 1, :].size, 1)
    xn = A.dot(x) + B.dot(un_1) + K

    if not pc.is_inside(P1, xn):
        inside = False

    return inside


def find_discrete_state(x0, part):
    """Return index identifying the discrete state
    to which the continuous state x0 belongs to.

    Notes
    =====
    1. If there are overlapping partitions
        (i.e., x0 belongs to more than one discrete state),
        then return the first discrete state ID

    @param x0: initial continuous state
    @type x0: numpy 1darray

    @param part: state space partition
    @type part: L{PropPreservingPartition}

    @return: if C{x0} belongs to some
        discrete state in C{part},
        then return the index of that state

        Otherwise return None, i.e., in case
        C{x0} does not belong to any discrete state.
    @rtype: int
    """
    for (i, region) in enumerate(part):
        if pc.is_inside(region, x0):
            return i
    return None



# Copyright (c) 2014, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Interface to Lily that solves LTL games.

Requires pydot for networkx to load the Moore strategy graph.

Relevant links:
  - U{Lily<http://www.ist.tugraz.at/staff/jobstmann/lily/>}
"""
import logging
import os
import re
import subprocess

import networkx as nx
import pydot

from tulip.spec.parser import parse
from tulip.spec.translation import translate
from tulip.spec.translation import translate_ast
from tulip.transys import MooreMachine
from tulip.spec.form import GRSpec
from tulip.spec.form import LTL


logger = logging.getLogger(__name__)
LILY = 'lily.pl'
IO_PARTITION_FILE = 'io_partition.txt'
DOTFILE = 'ltl2vl-synthesis.dot'


def synthesize(formula, env_vars=None, sys_vars=None):
    """Return Moore transducer if C{formula} is realizable.

    Variable C{dict}s have variable names as keys and their type as
    value. The types should be 'boolean'. These parameters are only
    used if formula is of type C{str}. Else, the variable dictionaries
    associated with the L{LTL} or L{GRSpec} object are used.

    @param formula: linear temporal logic formula
    @type formula: C{str}, L{LTL}, or L{GRSpec}

    @param env_vars: uncontrolled variables (inputs); used only if
        C{formula} is of type C{str}
    @type env_vars: C{dict} or None

    @param sys_vars: controlled variables (outputs); used only if
        C{formula} is of type C{str}
    @type sys_vars: C{dict} or None

    @return: symbolic Moore transducer
        (guards are conjunctions, not sets)
    @rtype: L{MooreMachine}
    """
    if isinstance(formula, GRSpec):
        env_vars = formula.env_vars
        sys_vars = formula.sys_vars
    elif isinstance(formula, LTL):
        env_vars = formula.input_variables
        sys_vars = formula.output_variables

    all_vars = dict(env_vars)
    all_vars.update(sys_vars)
    if not all(v == 'boolean' for v in all_vars.itervalues()):
        raise TypeError(
            'all variables should be Boolean:\n{v}'.format(v=all_vars))

    if isinstance(formula, GRSpec):
        f = translate(formula, 'wring')
    else:
        T = parse(str(formula))
        f = translate_ast(T, 'wring').flatten(env_vars=env_vars,
                                              sys_vars=sys_vars)

    # dump partition file
    s = '.inputs {inp}\n.outputs {out}'.format(
        inp=' '.join(env_vars),
        out=' '.join(sys_vars)
    )
    with open(IO_PARTITION_FILE, 'w') as fid:
        fid.write(s)

    # call lily
    try:
        p = subprocess.Popen([LILY, '-f', f, '-syn', IO_PARTITION_FILE],
                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        out = p.stdout.read()
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            raise Exception(
                'lily.pl not found in path.\n'
                'See the Lily docs for setting PERL5LIB and PATH.')
        else:
            raise

    dotf = open(DOTFILE, 'r')
    fail_msg = 'Formula is not realizable'
    if dotf.read(len(fail_msg)) == fail_msg:
        return None
    dotf.seek(0)
    data = dotf.read()
    (pd,) = pydot.graph_from_dot_data(data)
    g = nx.drawing.nx_pydot.from_pydot(pd)
    dotf.close()
    moore = lily_strategy2moore(g, env_vars, sys_vars)
    return moore


def lily_strategy2moore(g, env_vars, sys_vars):
    """Return Moore transducer from Lily strategy graph C{g}.

    Caution
    =======
    The resulting transducer is symboic,
    in that the guards denote conjunctions,
    *not* subsets of ports.

    @param g: Moore strategy game graph as output by Lily
    @type g: C{networkx.MultiDiGraph}

    @rtype: L{MooreMachine}
    """
    g.remove_node('title')
    phantom_init = {x for x in g if x.startswith('init-')}
    game_nodes = {x for x in g if x not in phantom_init}
    sys_nodes = {x for x in game_nodes if g.node[x].get('shape') is None}
    mapping = {k: i for i, k in enumerate(sys_nodes)}
    # avoid mapping MooreMachine, because it raises errors
    h = nx.relabel_nodes(g, mapping)

    m = MooreMachine()
    inports = {k: {False, True} for k in env_vars}
    outports = {k: {False, True} for k in sys_vars}
    m.add_inputs(inports)
    m.add_outputs(outports)
    m.add_nodes_from(mapping[x] for x in sys_nodes)

    # add initial states
    for u in phantom_init:
        for v in g.successors_iter(u):
            m.states.initial.add(mapping[v])

    # label vertices with output values
    for u in m:
        oute = h.out_edges(u, data=True)
        assert(len(oute) == 1)
        u_, v, attr = oute[0]
        assert(u_ is u)
        d = _parse_label(attr['label'])
        m.add_node(u, **d)

        # input doesn't matter for this reaction ?
        if v in m:
            m.add_edge(u, v)
            continue

        # label edges with input values that matter
        for v_, w, attr in h.out_edges_iter(v, data=True):
            assert(v_ is v)
            assert(w in m)
            d = _parse_label(attr['label'])
            m.add_edge(u, w, **d)
    return m


def _parse_label(s):
    """Return C{dict} from variable conjunction.

    @type s: C{str}

    @rtype: C{dict}
    """
    l = re.findall('(\w+)=(0|1)', s)
    return {k: bool(int(v)) for k, v in l}



# Copyright (c) 2013 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Finite State Machine examples

Bibliography:

[LS11] E.A. Lee and S.A. Seshia. *Introduction to Embedded Systems - A Cyber-
Physical Systems Approach*. `LeeSeshia.org <http://LeeSeshia.org>`_, 2011.
"""
import tulip.transys as trs
import tulip.transys.machines as mc

hl = 60*'='
save_fig = False

def mealy_machine_example():
    import numpy as np

    class check_diaphragm():
        """camera f-number."""
        def is_valid_value(x):
            if x <= 0.7 or x > 256:
                raise TypeError('This f-# is outside allowable range.')

        def __contains__(self, guard):
            # when properly implemented, do appropriate syntactic check
            if isinstance(guard, float):
                return True

            return False

        def __call__(self, guard_set, input_port_value):
            """This method "knows" that we are using x to denote the input
            within guards."""
            self.is_valid_value(input_port_value)

            guard_var_def = {'x': input_port_value}
            guard_value = eval(guard_set, guard_var_def)

            if not isinstance(guard_value, bool):
                raise TypeError('Guard value is non-boolean.\n'
                                'A guard is a predicate, '
                                'so it can take only boolean values.')

    class check_camera():
        """is it looking upwards ?"""
        def is_valid_value(self, x):
            if x.shape != (3,):
                raise Exception('Not a 3d vector!')

        def __contains__(self, guard):
            # when properly implemented, do appropriate syntactic check
            if isinstance(guard, np.ndarray) and guard.shape == (3,):
                return True

            return False

        def __call__(self, guard_set, input_port_value):
            self.is_valid_value(input_port_value)

            v1 = guard_set # guard_halfspace_normal_vector
            v2 = input_port_value # camera_direction_vector

            if np.inner(v1, v2) > 0.8:
                return True

            return False

    # note: guards are conjunctions,
    # any disjunction is represented by 2 edges

    # input defs
    inputs = {
        'speed': {'zero', 'low', 'high', 'crazy'},
        'seats': trs.PowerSet(range(5) ),
        'aperture': check_diaphragm(),
        'camera': check_camera()}

    # outputs def
    outputs = {'photo': {'capture', 'wait'}}

    # state variables def
    state_vars = {'light': {'on', 'off'}}

    # define the machine itself
    m = trs.MealyMachine()

    m.add_state_vars(state_vars)
    m.add_inputs(inputs)
    m.add_outputs(outputs)

    m.states.add('s0')
    m.states.add_from(['s1', 's2'])
    m.states.initial.add('s0')

    m.transitions.add(
        's0', 's1',
        speed='low',
        seats=(0, 1),
        aperture=0.3,
        camera=np.array([0,0,1]),
        photo='capture'
    )

    guard = {'camera':np.array([1,1,1]),
             'speed':'high',
             'aperture':0.3,
             'seats':(2, 3),
             'photo':'wait'}
    m.transitions.add('s1', 's2', **guard)

    m.plot(rankdir='TB')

    return m

def garage_counter(ploting=True):
    """Example 3.4, p.49 [LS11], for M=2

    no state variables in this Finite-State Machine
    """
    m = trs.MealyMachine()

    m.add_inputs({
        'up': {'present', 'absent'},
        'down': {'present', 'absent'}})

    m.add_outputs({'count': range(3)})

    m.states.add_from(range(3) )
    m.states.initial.add(0)

    m.transitions.add(0, 1, up='present', down='absent', count=1)
    m.transitions.add(1, 0, up='absent', down='present', count=0)

    m.transitions.add(1, 2, up='present', down='absent', count=2)
    m.transitions.add(2, 1, up='absent', down='present', count=1)

    if ploting:
        m.plot()

    return m

def garage_counter_with_state_vars():
    """Example 3.8, p.57 [LS11], for M=2
    unfolded with respect to state variable c
    """
    m = garage_counter(ploting=False)

    m.add_state_vars({'c': range(3)})
    m.states.add(0, c=0)
    m.states.add(1, c=1)
    m.states.add(2, c=2)

    m.save()

    return m

def thermostat_with_hysteresis():
    """Example 3.5, p.50 [LS11]
    """
    class temperature_type():
        def is_valid_value(x):
            if not isinstance(x, [float, int]):
                raise TypeError('Input temperature must be float.')

        def __contains__(self, guard):
            # when properly implemented, do appropriate syntactic check
            if isinstance(guard, float):
                return True

            return False

    m = trs.MealyMachine()

    m.add_inputs({'temperature': set()})

def traffic_light():
    m = trs.MealyMachine()
    pure_signal = {'present', 'absent'}

    m.add_inputs({'tick': pure_signal})
    m.add_outputs({'go': pure_signal, 'stop': pure_signal})

    m.states.add_from(['red', 'green', 'yellow'])
    m.states.initial.add('red')

    p = 'present'
    a = 'absent'

    m.transitions.add('red', 'green', tick=p, go=p, stop=a)
    m.transitions.add('green', 'yellow', tick=p, go=a, stop=p)
    m.transitions.add('yellow', 'red', tick=p, go=a, stop=p)

    m.save()
    return m

def pedestrians():
    """Example 2.14, p.63 [LS11]
    """
    m = trs.MealyMachine()

    m.add_inputs({
        'sigR': mc.pure,
        'sigG': mc.pure,
        'sigY': mc.pure})

    m.add_outputs({'pedestrian': mc.pure})

    m.states.add_from(['none', 'waiting', 'crossing'] )
    m.states.initial.add('crossing')

    for sigR in mc.pure:
        for sigG in mc.pure:
            for sigY in mc.pure:
                m.transitions.add(
                    'none', 'none',
                    sigR=sigR, sigG=sigG, sigY=sigY,
                    pedestrian='absent'
                )

                m.transitions.add(
                    'none', 'waiting',
                    sigR=sigR, sigG=sigG, sigY=sigY,
                    pedestrian='present'
                )

    m.transitions.add(
        'waiting', 'crossing',
        sigR='present', sigG='absent', sigY='absent',
        pedestrian='absent'
    )
    m.transitions.add(
        'crossing', 'none',
        sigR='absent', sigG='present', sigY='absent',
        pedestrian='absent'
    )
    m.save()
    return m

if __name__ == '__main__':
    saving = False

    m1 = mealy_machine_example()
    m2 = garage_counter()
    m3 = garage_counter_with_state_vars()
    m4 = pedestrians()
    m5 = traffic_light()

    m5.run(4)

    # save animated javascript
    if saving:
        m4.save('index.html', 'html')



#!/usr/bin/env python
"""Usage: genrand.py [-b] [-s] [-p] [H W]

will generate a random gridworld of the height H and width W (default
is 5 by 10) and dump the resulting description string.  An example
use-case is to save the result to a file for later fun by

  $ ./genrand.py 20 70 > goodtimes.txt

You could then load this in your code using, e.g,

  with open("goodtimes.txt", "r") as f:
      X = GridWorld(f.read())

If the flag "-s" is given, then a specification using TuLiP LTL syntax
is printed with the description string given in the comments.  If the
flag "-p" is given, then pretty-print the gridworld into the comments.

Use the flag -b to use a representation where there is one boolean
variable in the specification per grid cell.  Otherwise (default),
support for nonboolean domains is used.
"""
from __future__ import print_function

from os import environ as os_environ
import sys
import tulip.gridworld as gw

if os_environ.has_key("TULIP_REGRESS"):
    import numpy
    numpy.random.seed(0)


if len(sys.argv) > 5 or "-h" in sys.argv:
    print("Usage: genrand.py  [-s] [-p] [H W]")
    sys.exit(1)

if "-b" in sys.argv:
    nonbool = False
    sys.argv.remove("-b")
else:
    nonbool = True

if "-s" in sys.argv:
    dump_spec = True
    sys.argv.remove("-s")
else:
    dump_spec = False

if "-p" in sys.argv:
    print_pretty = True
    sys.argv.remove("-p")
else:
    print_pretty = False

if len(sys.argv) >= 3:
    (height, width) = (int(sys.argv[1]), int(sys.argv[2]))
else:
    (height, width) = (5, 10)

Z = gw.random_world((height, width),
                    wall_density=0.2,
                    num_init=1,
                    num_goals=2)

if dump_spec:
    if print_pretty:
        print(Z.pretty(show_grid=True, line_prefix="## "))
    print(Z.dumps(line_prefix="# "))
    print(Z.spec(nonbool=nonbool).pretty())
else:
    if print_pretty:
        print(Z.pretty(show_grid=True, line_prefix="# "))
    print(Z.dumps())



#!/usr/bin/env python
"""Tests for the export mechanisms of tulip.dumpsmach."""
import logging

import networkx as nx
from nose.tools import assert_raises

from tulip import spec, synth, dumpsmach


logging.getLogger('tulip').setLevel('ERROR')
logging.getLogger('astutils').setLevel('ERROR')
logging.getLogger('omega').setLevel('ERROR')


class basic_test:
    def setUp(self):
        self.triv = spec.GRSpec(env_vars="x", sys_vars="y",
                                env_init="x & y", env_prog="x",
                                sys_init="y", sys_prog="y && x")
        self.triv_M = synth.synthesize('omega', self.triv)

        self.dcounter = spec.GRSpec(
            sys_vars={"y": (0, 5)},
            env_init=['y = 0'],
            sys_prog=["y=0", "y=5"])
        self.dcounter_M = synth.synthesize('omega', self.dcounter)

        self.enumf = spec.GRSpec(
            sys_vars={'y': ['a', 'b']},
            env_init=['y="a"'],
            sys_safety=['y = "a" -> X(y = "b")',
                        'y = "b" -> X(y = "a")'])
        self.enumf_M = synth.synthesize('omega', self.enumf)

    def tearDown(self):
        self.dcounter = None
        self.dcounter_M = None

    def test_python_case(self):
        compile(dumpsmach.python_case(self.triv_M),
                filename="<string>", mode="exec")
        # print(dumpsmach.python_case(self.dcounter_M))
        compile(dumpsmach.python_case(self.dcounter_M),
                filename="<string>", mode="exec")
        exec compile(dumpsmach.python_case(self.enumf_M)
                     +'\nM = TulipStrategy(); M.move()',
                     filename="<string>", mode="exec")


def test_nx():
    g = nx.DiGraph()
    g.inputs = {'a': '...', 'b': '...'}
    g.outputs = {'c': '...', 'd': '...'}
    start = 'Sinit'
    g.add_edge(start, 0, a=0, b=0, c=0, d=0)
    g.add_edge(0, 1, a=0, b=1, c=0, d=1)
    g.add_edge(1, 2, a=1, b=0, c=1, d=1)
    print(dumpsmach.python_case(g, classname='Machine', start='Sinit'))
    exec dumpsmach.python_case(g, classname='Machine', start='Sinit')
    m = Machine()  # previous line creates the class `Machine`
    # Sinit -> 0
    out = m.move(a=0, b=0)
    assert out == dict(c=0, d=0)
    # 0 -> 1
    out = m.move(a=0, b=1)
    assert out == dict(c=0, d=1)
    # invalid input for index 2 in time sequence
    with assert_raises(ValueError):
        m.move(a=1, b=1)
    # 1 -> 2
    out = m.move(a=1, b=0)
    assert out == dict(c=1, d=1)
    # dead-end
    with assert_raises(Exception):
        m.move(a=1, b=0)



# Copyright (c) 2014, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
"""Test if given formula belongs to an LTL fragment that
is convertible to deterministic Buchi Automata
(readily expressible in GR(1) ).

reference
=========
1. Andreas Morgenstern and Klaus Schneider,
   A LTL Fragment for GR(1)-Synthesis,
   in Proceedings First International Workshop on
   Interactions, Games and Protocols (iWIGP),
   Electronic Proceedings in Theoretical Computer Science (EPTCS),
   50, pp. 33--45, 2011,
   http://doi.org/10.4204/EPTCS.50.3
"""
from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import networkx as nx
from tulip import transys as trs
from tulip.spec import lexyacc, GRSpec
from tulip.spec import transformation as tx
from tulip.spec import parser
from tulip.spec import ast as sast


def check(formula):
    """Parse formula string and create abstract syntax tree (AST).
    """
    ast = lexyacc.parse(formula)

    dfa = trs.automata.FiniteWordAutomaton(atomic_proposition_based=False,
                                           deterministic=True)

    dfa.alphabet |= {'!', 'W', 'U', 'G', 'F',
                     'U_left', 'U_right',
                     'W_left', 'W_right'}

    dfa.states.add_from({'gf', 'fg', 'g', 'f'})
    dfa.states.initial.add('gf')

    dfa.transitions.add('gf', 'fg', letter='!')
    dfa.transitions.add('fg', 'gf', letter='!')
    dfa.transitions.add('g', 'f', letter='!')
    dfa.transitions.add('f', 'g', letter='!')

    dfa.transitions.add('gf', 'gf', letter='W')
    dfa.transitions.add('gf', 'gf', letter='U_left')
    dfa.transitions.add('gf', 'gf', letter='G')

    dfa.transitions.add('fg', 'fg', letter='U')
    dfa.transitions.add('fg', 'fg', letter='F')
    dfa.transitions.add('fg', 'fg', letter='W_right')

    dfa.transitions.add('gf', 'f', letter='U_right')
    dfa.transitions.add('gf', 'f', letter='F')

    dfa.transitions.add('fg', 'g', letter='W_left')
    dfa.transitions.add('fg', 'g', letter='G')

    dfa.transitions.add('g', 'g', letter='W')
    dfa.transitions.add('g', 'g', letter='G')

    dfa.transitions.add('f', 'f', letter='U')
    dfa.transitions.add('f', 'f', letter='F')

    # plot tree automaton
    # dfa.save('dfa.pdf')

    # plot parse tree
    sast.dump_dot(ast, 'ast.dot')

    # sync product of AST with DFA,
    # to check acceptance
    Q = [(ast, 'gf')]
    while Q:
        s, q = Q.pop()
        logger.info('visiting: ' + str(s) + ', ' + str(q))

        if isinstance(s, sast.Unary):
            op = s.operator

            if op in {'!', 'G', 'F'}:
                t = dfa.transitions.find(q, letter=op)

                if not t:
                    raise Exception('not in fragment')

                qi, qj, w = t[0]

                Q.append((s.operand, qj))
            else:
                # ignore
                Q.append((s.operand, q))
        elif isinstance(s, sast.Binary):
            op = s.operator

            if op in {'W', 'U'}:
                t = dfa.transitions.find(q, letter=op)
                if t:
                    qi, qj, w = t[0]
                    Q.append((s.op_l, qj))
                    Q.append((s.op_r, qj))
                else:
                    t = dfa.transitions.find(q, letter=op + '_left')

                    if not t:
                        raise Exception('not in fragment')

                    qi, qj, w = t[0]

                    Q.append((s.op_l, qj))

                    t = dfa.transitions.find(q, letter=op + '_right')

                    if not t:
                        raise Exception('not in fragment')

                    qi, qj, w = t[0]

                    Q.append((s.op_r, qj))
            else:
                # ignore
                Q.append((s.op_l, q))
                Q.append((s.op_r, q))
        elif isinstance(s, sast.Var):
            print('reached var')

    return ast


def str_to_grspec(f):
    """Return `GRSpec` from LTL formula `f` as `str`.

    Formula `f` must be in the form:

      A -> G

    where each of A, G is a conjunction of terms: `B`, `[]C`, `[]<>B`.
    For more details on `B, C`, see [split_gr1].

    @type f: `str`
    @rtype: [GRSpec]
    """
    t = parser.parse(f)
    assert t.operator == '->'
    env, sys = t.operands
    d = {'assume': split_gr1(env),
         'assert': split_gr1(sys)}
    return GRSpec(env_init=d['assume']['init'],
                  env_safety=d['assume']['G'],
                  env_prog=d['assume']['GF'],
                  sys_init=d['assert']['init'],
                  sys_safety=d['assert']['G'],
                  sys_prog=d['assert']['GF'])


def split_gr1(f):
    """Return `dict` of GR(1) subformulae.

    The formula `f` is assumed to be a conjunction of expressions
    of the form:
      `B`, `[]C` or `[]<>B`
    where:
      - `C` can contain "next"
      - `B` cannot contain "next"

    @param f: temporal logic formula
    @type f: `str` or AST

    @return: conjunctions of formulae A, B as `str`, grouped by keys:
        `'init', 'G', 'GF'`
    @rtype: `dict` of `str`: `list` of `str`
    """
    # TODO: preprocess by applying syntactic identities: [][] = [] etc
    try:
        f + 's'
        t = parser.parse(f)
    except TypeError:
        t = f
    g = tx.Tree.from_recursive_ast(t)
    # collect boundary of conjunction operators
    Q = [g.root]
    b = list()  # use lists to preserve as much given syntactic order
    while Q:
        u = Q.pop()
        # terminal ?
        if not g.succ.get(u):
            b.append(u)
            continue
        # operator
        if u.operator == '&':
            # use `u.operands` instead of `g.successors`
            # to preserve original order
            Q.extend(u.operands)
        else:
            b.append(u)
    d = {'init': list(), 'G': list(), 'GF': list()}
    for u in b:
        # terminal ?
        if not g.succ.get(u):
            d['init'].append(u)
            continue
        # some operator
        if u.operator != 'G':
            d['init'].append(u)
            continue
        # G
        (v,) = u.operands
        # terminal in G ?
        if not g.succ.get(v):
            d['G'].append(v)
            continue
        # some operator in G
        if v.operator == 'F':
            (w,) = v.operands
            d['GF'].append(w)
        else:
            # not a GF
            d['G'].append(v)
    # assert only admissible temporal operators
    ops = {'G', 'F', 'U', 'V', 'R'}
    operators = {'G': ops}
    ops = set(ops)
    ops.add('X')
    operators.update(init=ops, GF=ops)
    for part, f in d.iteritems():
        ops = operators[part]
        for u in f:
            op = has_operator(u, g, ops)
            if op is None:
                continue
            raise AssertionError((
                'found inadmissible operator "{op}" '
                'in "{f}" formula').format(
                    op=op, f=u))
    # conjoin (except for progress)
    init = ' & '.join(u.flatten() for u in reversed(d['init']))
    d['init'] = [init]
    safe = ' & '.join(u.flatten() for u in reversed(d['G']))
    d['G'] = [safe]
    # flatten individual progress formulae
    d['GF'] = [u.flatten() for u in d['GF']]
    return d


def has_operator(u, g, operators):
    try:
        if u.operator in operators:
            return u.operator
    except AttributeError:
        pass
    for v in nx.descendants(g, u):
        # terminal
        if not g.succ.get(v):
            continue
        # operator
        # is it temporal except for 'X' ?
        if v.operator in operators:
            return v.operator
    return None


def stability_to_gr1(p, aux='aux'):
    """Convert C{<>[] p} to GR(1).

    Warning: This conversion is sound, but not complete.
    See p.2, U{[E10]
    <http://tulip-control.sourceforge.net/doc/bibliography.html#e10>}

    GR(1) form::

        !(aux) &&
        [](aux -> X aux) &&
        []<>(aux) &&

        [](aux -> p)

    @type p: str

    @param aux: name to use for auxiliary variable
    @type aux: str

    @rtype: L{GRSpec}
    """
    logging.warning(
        'Conversion of stability (<>[]p) to GR(1)' +
        'is sound, but NOT complete.'
    )

    a = aux
    a0 = a

    p = _paren(p)
    a = _paren(a)

    v = tx.check_var_name_conflict(p, a0)

    sys_vars = v | {a0}
    sys_init = {'!' + a}
    sys_safe = {a + ' -> ' + p,
                a + ' -> X ' + a}
    sys_prog = {a}

    return GRSpec(sys_vars=sys_vars, sys_init=sys_init,
                  sys_safety=sys_safe, sys_prog=sys_prog)


def response_to_gr1(p, q, aux='aux'):
    """Convert C{[](p -> <> q)} to GR(1).

    GR(1) form::

        []<>(aux) &&

        []( (p && !q) -> X ! aux) &&
        []( (! aux && !q) -> X ! aux)

    @type p: str

    @type q: str

    @param aux: name to use for auxiliary variable
    @type aux: str

    @rtype: L{GRSpec}
    """
    a = aux
    a0 = a

    p = _paren(p)
    q = _paren(q)
    a = _paren(a)

    s = p + ' -> <> ' + q
    v = tx.check_var_name_conflict(s, a0)

    sys_vars = v | {a0}
    # sys_init = {a}
    sys_safe = {
        '(' + p + ' && !' + q + ') -> X !' + a,
        '(!' + a + ' && !' + q + ') -> X !' + a
    }
    sys_prog = {a}

    return GRSpec(sys_vars=sys_vars,  # sys_init=sys_init,
                  sys_safety=sys_safe, sys_prog=sys_prog)


def eventually_to_gr1(p, aux='aux'):
    """Convert C{<> p} to GR(1).

    GR(1) form::

        !(aux) &&
        [](aux -> X aux) &&
        []<>(aux) &&

        []( (!p && !aux) -> X!(aux) )

    @type p: str

    @param aux: name to use for auxiliary variable
    @type aux: str

    @rtype: L{GRSpec}
    """
    a = aux
    a0 = a

    p = _paren(p)
    a = _paren(a)

    v = tx.check_var_name_conflict(p, a0)

    sys_vars = v | {a0}
    sys_init = {'!(' + a + ')'}
    sys_safe = {
        '(!' + p + ' && !' + a + ') -> X !' + a,
        a + ' -> X ' + a
    }
    sys_prog = {a}

    return GRSpec(sys_vars=sys_vars, sys_init=sys_init,
                  sys_safety=sys_safe, sys_prog=sys_prog)


def until_to_gr1(p, q, aux='aux'):
    """Convert C{p U q} to GR(1).

    GR(1) form::

        (!q -> !aux) &&
        [](q -> aux)
        [](aux -> X aux) &&
        []<>(aux) &&

        []( (!aux && X(!q) ) -> X!(aux) ) &&
        [](!aux -> p)

    @type p: str

    @param aux: name to use for auxiliary variable
    @type aux: str

    @rtype: L{GRSpec}
    """
    a = aux
    a0 = a

    p = _paren(p)
    q = _paren(q)
    a = _paren(a)

    s = p + ' && ' + q
    v = tx.check_var_name_conflict(s, a0)

    sys_vars = v | {a0}
    sys_init = {'!' + q + ' -> !' + a}
    sys_safe = {
        q + ' -> ' + a,
        '( (X !' + q + ') && !' + a + ') -> X !' + a,
        a + ' -> X ' + a,
        '(!' + a + ') -> ' + p
    }
    sys_prog = {a}

    return GRSpec(sys_vars=sys_vars, sys_init=sys_init,
                  sys_safety=sys_safe, sys_prog=sys_prog)


def _paren(x):
    return '({x})'.format(x=x)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    s = '(a U b) && []a && <>a && <>a && []<>(<>z)'
    parsed_formula = check(s)

    print('Parsing result: ' + str(parsed_formula))
